# -*- coding: utf-8 -*-
""" Handles reading and interpolation of atmopshere (grid) data """
import itertools
import logging

import numpy as np
from astropy import constants as const
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit

from ..large_file_storage import setup_atmo
from .atmosphere import Atmosphere as Atmo
from .atmosphere import AtmosphereError, AtmosphereGrid
from .savfile import SavFile

logger = logging.getLogger(__name__)

# Radius and Surface Gravity of the Sun
# Required for spherical models
R_sun = const.R_sun.to_value("cm")
g_sun = (const.G * const.M_sun / const.R_sun ** 2).to_value("cm/s**2")
logg_sun = np.log10(g_sun)


class AtmosphereInterpolator:
    def __init__(self, depth=None, interp=None, geom=None, lfs_atmo=None, verbose=0):
        self.depth = depth
        self.interp = interp
        self.geom = geom
        if lfs_atmo is None:
            lfs_atmo = setup_atmo()
        self.lfs_atmo = lfs_atmo

        self.source = None
        self.atmo_grid = None
        self.verbose = verbose

    def interp_atmo_grid(self, atmo_grid, teff, logg, monh):
        """
        General routine to interpolate in 3D grid of model atmospheres

        Parameters
        -----
        Teff : float
            effective temperature of desired model (K).
        logg : float
            logarithmic gravity of desired model (log cm/s/s).
        MonH : float
            metalicity of desired model.
        atmo_in : Atmosphere
            Input atmosphere
        verbose : {0, 1}, optional
            how much information to plot/print (default: 0)
        plot : bool, optional
            wether to plot debug information (default: False)
        reload : bool
            wether to reload atmosphere information from disk (default: False)

        Returns
        -------
        atmo : Atmosphere
            interpolated atmosphere data
        """

        # Internal parameters.
        if not isinstance(atmo_grid, AtmosphereGrid):
            if self.atmo_grid is None or self.source != atmo_grid:
                atmo_file = self.lfs_atmo.get(atmo_grid)
                self.source = atmo_grid
                self.atmo_grid = SavFile(
                    atmo_file, source=self.source, lfs=self.lfs_atmo
                )
        else:
            self.atmo_grid = atmo_grid
            self.source = atmo_grid.source

        if self.geom is not None:
            if self.geom == "PP":
                atmo_grid = self.atmo_grid[self.atmo_grid.radius <= 1]
            elif self.geom == "SPH":
                atmo_grid = self.atmo_grid[self.atmo_grid.radius > 1]
        else:
            atmo_grid = self.atmo_grid

        assert len(atmo_grid) > 0, "No atmospheres for this geometry"

        # Get field names in ATMO and ATMO_GRID structures.
        depth = self.determine_depth_scale(self.depth, atmo_grid)
        interp = self.determine_interpolation_scale(self.interp, atmo_grid)

        # Find the corner models bracketing the given values
        icor = self.find_corner_models(teff, logg, monh, atmo_grid)

        # Interpolate the corner models
        atmo = self.interpolate_corner_models(
            teff, logg, monh, icor, atmo_grid, interp=interp
        )

        # TODO: Or should we only consider spherical models for interpolation of spherical modesl are requested?
        geom, radius = self.spherical_model_correction(atmo_grid, icor, logg)
        # Create ATMO.GEOM, if necessary, and set value.
        if self.geom is not None and self.geom != geom:
            if self.geom == "SPH":
                raise AtmosphereError(
                    "Input ATMO.GEOM='%s' was requested but the model only supports PP (at this point)."
                    % self.geom
                )
            else:
                logger.info(
                    "Input ATMO.GEOM='%s' overrides '%s' from grid.",
                    self.geom,
                    geom,
                )

        # Add standard ATMO input fields, if they are missing from ATMO_IN.
        atmo.depth = depth
        atmo.interp = interp
        atmo.geom = geom
        if radius is not None:
            atmo.radius = radius
        atmo.source = self.source
        atmo.method = "grid"

        return atmo

    def interp_atmo_pair(self, atmo1, atmo2, frac, interpvar="RHOX", itop=0):
        """
        Interpolate between two model atmospheres, accounting for shifts in
        the mass column density or optical depth scale.

        How it works:

        1. The second atmosphere is fitted onto the first, individually for
        each of the four atmospheric quantitites: T, xne, xna, rho.
        The fitting uses a linear shift in both the (log) depth parameter and
        in the (log) atmospheric quantity. For T, the midpoint of the two
        atmospheres is aligned for the initial guess. The result of this fit
        is used as initial guess for the other quantities. A penalty function
        is applied to each fit, to avoid excessively large shifts on the
        depth scale.
        2. The mean of the horizontal shift in each parameter is used to
        construct the common output depth scale.
        3. Each atmospheric quantity is interpolated after shifting the two
        corner models by the amount determined in step 1), rescaled by the
        interpolation fraction (frac).

        Parameters
        ------
        atmo1 : Atmosphere
            first atmosphere to interpolate
        atmo2 : Atmosphere
            second atmosphere to interpolate
        frac : float
            interpolation fraction: 0.0 -> atmo1 and 1.0 -> atmo2
        interpvar : {"RHOX", "TAU"}, optional
            atmosphere interpolation variable (default:"RHOX").
        itop : int, optional
            index of top point in the atmosphere to use. default
            is to use all points (itop=0). use itop=1 to clip top depth point.
        atmop : array[5, ndep], optional
            interpolated atmosphere prediction (for plots)
            Not needed if atmospheres are provided as structures. (default: None)
        verbose : {0, 5}, optional
            diagnostic print level (default 0: no printing)
        plot : {-1, 0, 1}, optional
            diagnostic plot level. Larger absolute
            value yields more plots. Negative values cause a wait for keypress
            after each plot. (default: 0, no plots)
        old : bool, optional
            also plot result from the old interpkrz2 algorithm. (default: False)

        Returns
        ------
        atmo : Atmosphere
            interpolated atmosphere
            .RHOX (vector[ndep]) mass column density (g/cm^2)
            .TAU  (vector[ndep]) reference optical depth (at 5000 Å)
            .temp (vector[ndep]) temperature (K)
            .xne  (vector[ndep]) electron number density (1/cm^3)
            .xna  (vector[ndep]) atomic number density (1/cm^3)
            .rho  (vector[ndep]) mass density (g/cm^3)
        """
        # """
        # History
        # -------
        # 2004-Apr-15 Valenti
        #     Initial coding.
        # MB
        #     interpolation on TAU scale
        # 2012-Oct-30 TN
        #     Rewritten to use either column mass (RHOX) or
        #     reference optical depth (TAU) as vertical scale. Shift-interpolation
        #     algorithms have been improved for stability in cool dwarfs (<=3500 K).
        #     The reference optical depth scale is preferred in terms of interpolation
        #     accuracy across most of parameter space, with significant improvement for
        #     both cool models (where depth vs temperature is rather flat) and hot
        #     models (where depth vs temperature exhibits steep transitions).
        #     Column mass depth is used by default for backward compatibility.
        # 2013-May-17 Valenti
        #     Use frac to weight the two shifted depth scales,
        #     rather than simply averaging them. This change fixes discontinuities
        #     when crossing grid nodes.
        # 2013-Sep-10 Valenti
        #     Now returns an atmosphere structure instead of a
        #     [5,NDEP] atmosphere array. This was necessary to support interpolation
        #     using one variable (e.g., TAU) and radiative transfer using a different
        #     variable (e.g. RHOX). The atmosphere array could only store one depth
        #     variable, meaning the radiative transfer variable had to be the same
        #     as the interpolation variable. Returns atmo.RHOX if available and also
        #     atmo.TAU if available. Since both depth variables are returned, if
        #     available, this routine no longer needs to know which depth variable
        #     will be used for radiative transfer. Only the interpolation variable
        #     is important. Thus, the interpvar= keyword argument replaces the
        #     type= keyword argument. Very similar code blocks for each atmospheric
        #     quantity have been unified into a single code block inside a loop over
        #     atmospheric quantities.
        # 2013-Sep-21 Valenti
        #     Fixed an indexing bug that affected the output depth
        #     scale but not other atmosphere vectors. Itop clipping was not being
        #     applied to the depth scale ('RHOX' or 'TAU'). Bug fixed by adding
        #     interpvar to vtags. Now atmospheres interpolated with interp_atmo_grid
        #     match output from revision 398. Revisions back to 399 were development
        #     only, so no users should be affected.
        # 2014-Mar-05 Piskunov
        #     Replicated the removal of the bad top layers in models
        #     for interpvar eq 'TAU'
        # """

        # Internal program parameters.
        min_drhox = min_dtau = 0.01  # minimum fractional step in RHOX
        # min_dtau = 0.01  # minimum fractional step in TAU
        interpvar = interpvar.lower()
        ##
        ## Select interpolation variable (RHOX vs. TAU)
        ##

        # Check which depth scales are available in both input atmospheres.
        tags1 = atmo1.dtype.names
        tags2 = atmo2.dtype.names
        ok_tau = "tau" in tags1 and "tau" in tags2
        ok_rhox = "rhox" in tags1 and "rhox" in tags2
        if not ok_tau and not ok_rhox:
            raise AtmosphereError(
                "atmo1 and atmo2 structures must both contain RHOX or TAU"
            )

        # Set interpolation variable, if not specified by keyword argument.
        if interpvar is None:
            interpvar = "tau" if ok_tau else "rhox"
        if interpvar != "tau" and interpvar != "rhox":
            raise AtmosphereError("interpvar must be 'TAU' (default) or 'RHOX'")

        ##
        ## Define depth scale for both atmospheres
        ##

        # Define depth scale for atmosphere #1
        mask1 = np.full(len(atmo1[interpvar]), True)
        mask1[:itop] = False

        itop1 = itop
        while (atmo1[interpvar][itop1 + 1] / atmo1[interpvar][itop1] - 1) <= min_drhox:
            mask1[itop1] = False
            itop1 += 1

        mask1 &= atmo1[interpvar] != 0
        mask1 &= np.isfinite(atmo1[interpvar])
        ndep1 = np.count_nonzero(mask1)

        depth1 = np.log10(atmo1[interpvar][mask1])

        # Define depth scale for atmosphere #2
        mask2 = np.full(len(atmo2[interpvar]), True)
        mask2[:itop] = False
        itop2 = itop
        while (atmo2[interpvar][itop2 + 1] / atmo2[interpvar][itop2] - 1) <= min_drhox:
            mask2[itop2] = False
            itop2 += 1

        mask2 &= atmo2[interpvar] != 0
        mask2 &= np.isfinite(atmo2[interpvar])
        ndep2 = np.count_nonzero(mask2)

        depth2 = np.log10(atmo2[interpvar][mask2])

        ##
        ## Prepare to find best shift parameters for each atmosphere vector.
        ##

        # List of atmosphere vectors that need to be shifted.
        # The code below assumes 'TEMP' is the first vtag in the list.
        vtags = ["temp", "xne", "xna", "rho", interpvar]
        if interpvar == "rhox" and ok_tau:
            vtags += ["tau"]
        if interpvar == "tau" and ok_rhox:
            vtags += ["rhox"]
        nvtag = len(vtags)

        # Adopt arbitrary uncertainties for shift determinations.
        err1 = np.full(ndep1, 0.05)

        # Initial guess for TEMP shift parameters.
        # Put depth and TEMP midpoints for atmo1 and atmo2 on top of one another.
        npar = 4
        ipar = np.zeros(npar, dtype="f4")
        temp1 = np.log10(atmo1.temp[mask1])
        temp2 = np.log10(atmo2.temp[mask2])
        mid1 = np.argmin(np.abs(temp1 - 0.5 * (temp1[1] + temp1[-2])))
        mid2 = np.argmin(np.abs(temp2 - 0.5 * (temp2[1] + temp2[-2])))
        ipar[0] = depth1[mid1] - depth2[mid2]  # horizontal
        ipar[1] = temp1[mid1] - temp2[mid2]  # vertical

        # Apply a constraint on the fit, to avoid runaway for cool models, where
        # the temperature structure is nearly linear with both TAU and RHOX.
        constraints = np.zeros(npar)
        constraints[0] = 0.5  # weakly constrain the horizontal shift

        # For first pass ('TEMP'), use all available depth points.
        igd = np.isfinite(depth1)
        ngd = igd.size

        ##
        ## Find best shift parameters for each atmosphere vector.
        ##

        # Loop through atmosphere vectors.
        pars = np.zeros((nvtag, npar))
        for ivtag, vtag in enumerate(vtags):

            # Find vector in each structure.
            if vtag not in tags1:
                raise AtmosphereError("atmo1 does not contain " + vtag)
            if vtag not in tags2:
                raise AtmosphereError("atmo2 does not contain " + vtag)

            vect1 = np.log10(atmo1[vtag][mask1])
            vect2 = np.log10(atmo2[vtag][mask2])

            # Fit the second atmosphere onto the first by finding the best horizontal
            # shift in depth2 and the best vertical shift in vect2.
            pars[ivtag], _ = self.interp_atmo_constrained(
                depth1[igd],
                vect1[igd],
                err1[igd],
                ipar,
                x2=depth2,
                y2=vect2,
                y1=vect1,
                ndep=ngd,
                constraints=constraints,
            )

            # After first pass ('TEMP'), adjust initial guess and restrict depth points.
            if ivtag == 0:
                ipar = [pars[0, 0], 0.0, 0.0, 0.0]
                igd = np.where(
                    (depth1 >= min(depth2[igd]) + ipar[0])
                    & (depth1 <= max(depth2[igd]) + ipar[0])
                )[0]
                if igd.size < 2:
                    raise AtmosphereError("unstable shift in temperature")

        ##
        ## Use mean shift to construct output depth scale.
        ##

        # Calculate the mean depth2 shift for all atmosphere vectors.
        xsh = np.sum(pars[:, 0]) / nvtag

        # Base the output depth scale on the input scale with the fewest depth points.
        # Combine the two input scales, if they have the same number of depth points.
        depth1f = depth1 - xsh * frac
        depth2f = depth2 + xsh * (1 - frac)
        if ndep1 > ndep2:
            depth = depth2f
        elif ndep1 == ndep2:
            depth = depth1f * (1 - frac) + depth2f * frac
        elif ndep1 < ndep2:
            depth = depth1f
        ndep = len(depth)

        ##
        ## Interpolate input atmosphere vectors onto output depth scale.
        ##

        # Loop through atmosphere vectors.
        vects = np.zeros((nvtag, ndep))
        for ivtag, (vtag, par) in enumerate(zip(vtags, pars)):

            # Extract data
            vect1 = np.log10(atmo1[vtag][mask1])
            vect2 = np.log10(atmo2[vtag][mask2])

            # Identify output depth points that require extrapolation of atmosphere vector.
            depth1f = depth1 - par[0] * frac
            depth2f = depth2 + par[0] * (1 - frac)
            x1max = np.max(depth1f)
            x2max = np.max(depth2f)
            iup = (depth > x1max) | (depth > x2max)
            nup = np.count_nonzero(iup)
            checkup = (nup >= 1) and abs(frac - 0.5) <= 0.5 and ndep1 == ndep2

            # Combine shifted vect1 and vect2 structures to get output vect.
            vect1f = self.interp_atmo_func(depth, -frac * par, x2=depth1, y2=vect1)
            vect2f = self.interp_atmo_func(depth, (1 - frac) * par, x2=depth2, y2=vect2)
            vect = (1 - frac) * vect1f + frac * vect2f
            ends = [vect1[ndep1 - 1], vect[ndep - 1], vect2[ndep2 - 1]]
            if (
                checkup
                and np.median(ends) != vect[ndep - 1]
                and (
                    abs(vect1[ndep1 - 1] - 4.2) < 0.1
                    or abs(vect2[ndep2 - 1] - 4.2) < 0.1
                )
            ):
                vect[iup] = vect2f[iup] if x1max < x2max else vect1f[iup]
            vects[ivtag] = vect

        ##
        ## Construct output structure
        ##

        # Construct output structure with interpolated atmosphere.
        # Might be wise to interpolate abundances, in case those ever change.
        atmo = Atmo(interp=interpvar)
        stags = ["teff", "logg", "monh", "vturb", "lonh", "abund"]
        ndep_orig = len(atmo1.temp)
        for tag in tags1:

            # Default is to copy value from atmo1. Trim vectors.
            value = atmo1[tag]
            if np.size(value) == ndep_orig and tag != "abund":
                value = value[:ndep]

            # Vector quantities that have already been interpolated.
            if tag in vtags:
                ivtag = [i for i in range(nvtag) if tag == vtags[i]][0]
                value = 10.0 ** vects[ivtag]

            # Scalar quantities that should be interpolated using frac.
            if tag in stags:
                if tag in tags2:
                    value = (1 - frac) * atmo1[tag] + frac * atmo2[tag]
                else:
                    value = atmo1[tag]

            # Remaining cases.
            if tag == "ndep":
                value = ndep

            # Abundances
            if tag == "abund":
                value = (1 - frac) * atmo1[tag] + frac * atmo2[tag]

            # Create or add to output structure.
            atmo[tag] = value
        return atmo

    def determine_depth_scale(self, depth, atmo_grid):
        """
        Determine ATMO.DEPTH radiative transfer depth variable. Order of precedence:

        1. Value of ATMO_IN.DEPTH, if it exists and is set
        2. Value of ATMO_GRID[0].DEPTH, if it exists and is set
        3. 'RHOX', if ATMO_GRID.RHOX exists (preferred over 'TAU' for depth)
        4. 'TAU', if ATMO_GRID.TAU exists

        Check that INTERP is valid and the corresponding field exists in ATMO.

        Parameters
        ----------
        depth : {"RHOX", "TAU", None}
            requested value, or None for autoselection based in available grid
        atmo_grid : AtmosphereGrid
            input atmosphere grid to interpolate on

        Returns
        -------
        depth : {"TAU", "RHOX"}
            The chosen depth scale

        Raises
        ------
        AtmosphereError
            If an invalid value was set in atmo_in/atmo_grid
        """

        gtags = atmo_grid.dtype.names

        if depth is not None:
            depth = depth
        elif "depth" in gtags and atmo_grid.depth is not None:
            depth = atmo_grid.depth
        elif "rhox" in gtags:
            depth = "RHOX"
        elif "tau" in gtags:
            depth = "TAU"
        else:
            raise AtmosphereError("no value for ATMO.DEPTH")
        if depth != "TAU" and depth != "RHOX":
            raise AtmosphereError(
                "ATMO.DEPTH must be 'TAU' or 'RHOX', not '%s'" % depth
            )
        if depth.lower() not in gtags:
            raise AtmosphereError(
                "ATMO.DEPTH='{}', but ATMO. {} does not exist".format(depth, depth)
            )
        return depth

    def determine_interpolation_scale(self, interp, atmo_grid):
        """
        Determine ATMO.INTERP interpolation variable. Order of precedence:

        1. Value of ATMO_IN.INTERP, if it exists and is set
        2. Value of ATMO_GRID[0].INTERP, if it exists and is set
        3. 'TAU', if ATMO_GRID.TAU exists (preferred over 'RHOX' for interpolation)
        4. 'RHOX', if ATMO_GRID.RHOX exists

        Check that INTERP is valid and the corresponding field exists in ATMO.

        Parameters
        ----------
        interp : {"RHOX", "TAU", None}
            requested interpolation axis, or None for autoselect
        atmo_grid : AtmosphereGrid
            Atmosphere grid for interpolation

        Returns
        -------
        interp: {"RHOX", "TAU"}
            the interpolation axis

        Raises
        ------
        AtmosphereError
            if a non valid parameter is set in atmo_in/atmo_grid
        """

        gtags = atmo_grid.dtype.names

        if interp is not None:
            interp = interp
        elif atmo_grid.interp is not None:
            interp = atmo_grid.interp
        elif "tau" in gtags:
            interp = "TAU"
        elif "rhox" in gtags:
            interp = "RHOX"
        else:
            raise AtmosphereError("no value for ATMO.INTERP")
        if interp not in ["TAU", "RHOX"]:
            raise AtmosphereError(
                "ATMO.INTERP must be 'TAU' or 'RHOX', not '%s'" % interp
            )
        if interp.lower() not in gtags:
            raise AtmosphereError(
                "ATMO.INTERP='{}', but ATMO. {} does not exist".format(interp, interp)
            )
        return interp

    def find_corner_models(self, teff, logg, monh, atmo_grid):
        """
        Find the models in the grid that bracket the given stellar parameters

        The purpose of the first major set of code blocks is to find values
        of [M/H] in the grid that bracket the requested [M/H]. Then in this
        subset of models, find values of log(g) in the subgrid that bracket
        the requested log(g). Then in this subset of models, find values of
        Teff in the subgrid that bracket the requested Teff. The net result
        is a set of 8 models in the grid that bracket the requested stellar
        parameters. Only these 8 "corner" models will be used in the
        interpolation that constitutes the remainder of the program.

        Parameters
        ----------
        teff : float
            effective temperature
        logg : float
            surface gravity
        monh : float
            metallicity
        atmo_grid : AtmosphereGrid
            atmosphere grid to search models in
        verbose : int, optional
            determines how many debugging messages to print, by default 0
        """

        nb = 2  # number of bracket points

        # *** DETERMINATION OF METALICITY BRACKET ***
        # Find unique set of [M/H] values in grid.
        mlist = np.unique(atmo_grid.monh)  # list of unique [M/H]

        # Test whether requested metalicity is in grid.
        mmin = np.min(mlist)  # range of [M/H] in grid
        mmax = np.max(mlist)
        if monh > mmax:  # true: [M/H] too large
            logger.info(
                "interp_atmo_grid: requested [M/H] (%.3f) larger than max grid value (%.3f). extrapolating.",
                monh,
                mmax,
            )
        if monh < mmin:  # true: logg too small
            raise AtmosphereError(
                "interp_atmo_grid: requested [M/H] (%.3f) smaller than min grid value (%.3f). returning."
                % (monh, mmin)
            )

        # Find closest two [M/H] values in grid that bracket requested [M/H].
        if monh <= mmax:
            mlo = np.max(mlist[mlist <= monh])
            mup = np.min(mlist[mlist >= monh])
        else:
            mup = mmax
            mlo = np.max(mlist[mlist < mup])
        mb = [mlo, mup]  # bounding [M/H] values

        # Trace diagnostics.
        if self.verbose >= 5:
            logger.info("[M/H]: %.3f < %.3f < %.3f", mlo, monh, mup)

        # *** DETERMINATION OF LOG(G) BRACKETS AT [M/H] BRACKET VALUES ***
        # Set up for loop through [M/H] bounds.
        gb = np.zeros((nb, nb))  # bounding gravities
        for i in range(nb):
            # Find unique set of gravities at boundary below [M/H] value.
            im = atmo_grid.monh == mb[i]  # models on [M/H] boundary
            glist = np.unique(atmo_grid[im].logg)  # list of unique gravities

            # Test whether requested logarithmic gravity is in grid.
            gmin = np.min(glist)  # range of gravities in grid
            gmax = np.max(glist)
            if logg > gmax:  # true: logg too large
                logger.info(
                    "interp_atmo_grid: requested log(g) (%.3f) larger than max grid value (%.3f). extrapolating.",
                    logg,
                    gmax,
                )

            if logg < gmin:  # true: logg too small
                raise AtmosphereError(
                    "interp_atmo_grid: requested log(g) (%.3f) smaller than min grid value (%.3f). returning."
                    % (logg, gmin)
                )

            # Find closest two gravities in Mlo subgrid that bracket requested gravity.
            if logg <= gmax:
                glo = np.max(glist[glist <= logg])
                gup = np.min(glist[glist >= logg])
            else:
                gup = gmax
                glo = np.max(glist[glist < gup])
            gb[i] = [glo, gup]  # store boundary values.

            # Trace diagnostics.
            if self.verbose >= 5:
                logger.info(
                    "log(g) at [M/H]=%.3f: %.3f < %.3f < %.3f",
                    mb[i],
                    glo,
                    logg,
                    gup,
                )

        # End of loop through [M/H] bracket values.
        # *** DETERMINATION OF TEFF BRACKETS AT [M/H] and LOG(G) BRACKET VALUES ***
        # Set up for loop through [M/H] and log(g) bounds.
        tb = np.zeros((nb, nb, nb))  # bounding temperatures
        for ig in range(nb):
            for im in range(nb):
                # Find unique set of gravities at boundary below [M/H] value.
                it = (atmo_grid.monh == mb[im]) & (
                    atmo_grid.logg == gb[im, ig]
                )  # models on joint boundary
                tlist = np.unique(atmo_grid[it].teff)  # list of unique temperatures

                # Test whether requested temperature is in grid.
                tmin = np.min(tlist)  # range of temperatures in grid
                tmax = np.max(tlist)
                if teff > tmax:  # true: Teff too large
                    raise AtmosphereError(
                        "interp_atmo_grid: requested Teff (%i) larger than max grid value (%i). returning."
                        % (teff, tmax)
                    )
                if teff < tmin:  # true: logg too small
                    logger.info(
                        "interp_atmo_grid: requested Teff (%i) smaller than min grid value (%i). extrapolating.",
                        teff,
                        tmin,
                    )

                # Find closest two temperatures in subgrid that bracket requested Teff.
                if teff > tmin:
                    tlo = np.max(tlist[tlist <= teff])
                    tup = np.min(tlist[tlist >= teff])
                else:
                    tlo = tmin
                    tup = np.min(tlist[tlist > tlo])
                tb[im, ig, :] = [tlo, tup]  # store boundary values.

                # Trace diagnostics.
                if self.verbose >= 5:
                    logger.info(
                        "Teff at log(g)=%.3f and [M/H]=%.3f: %i < %i < %i",
                        gb[im, ig],
                        mb[im],
                        tlo,
                        teff,
                        tup,
                    )

        # End of loop through log(g) and [M/H] bracket values.

        # Find and save atmo_grid indices for the 8 corner models.
        icor = np.zeros((nb, nb, nb), dtype=int)
        for it, ig, im in itertools.product(range(nb), repeat=3):
            iwhr = np.where(
                (atmo_grid.teff == tb[im, ig, it])
                & (atmo_grid.logg == gb[im, ig])
                & (atmo_grid.monh == mb[im])
            )[0]
            nwhr = iwhr.size
            if nwhr != 1:
                logger.info(
                    "interp_atmo_grid: %i models in grid with [M/H]=%.1f, log(g)=%.1f, and Teff=%i",
                    nwhr,
                    mb[im],
                    gb[im, ig],
                    tb[im, ig, it],
                )
            icor[im, ig, it] = iwhr[0]

        # Trace diagnostics.
        if self.verbose >= 1:
            logger.info("Teff=%i,  log(g)=%.3f,  [M/H]=%.3f:", teff, logg, monh)
            logger.info("indx  M/H  g   Teff     indx  M/H  g   Teff")
            for im in range(nb):
                for ig in range(nb):
                    i0 = icor[im, ig, 0]
                    i1 = icor[im, ig, 1]
                    logger.info(
                        i0,
                        atmo_grid[i0].monh,
                        atmo_grid[i0].logg,
                        atmo_grid[i0].teff,
                        i1,
                        atmo_grid[i1].monh,
                        atmo_grid[i1].logg,
                        atmo_grid[i1].teff,
                    )
        return icor

    def interpolate_corner_models(
        self, teff, logg, monh, icor, atmo_grid, interp="RHOX", itop=1
    ):
        """
        Interpolate over the 8 corner models in the cube around the stellar parameters

        The code below interpolates between 8 corner models to produce
        the output atmosphere. In the first step, pairs of models at each
        of the 4 combinations of log(g) and Teff are interpolated to the
        desired value of [M/H]. These 4 new models are then interpolated
        to the desired value of log(g), yielding 2 models at the requested
        [M/H] and log(g). Finally, this pair of models is interpolated
        to the desired Teff, producing a single output model.

        Interpolation is done on the logarithm of all quantities to improve
        linearity of the fitted data. Kurucz models sometimes have very small
        fractional steps in mass column at the top of the atmosphere. These
        cause wild oscillations in splines fitted to facilitate interpolation
        onto a common depth scale. To circumvent this problem, we ignore the
        top point in the atmosphere by setting itop=1.

        Parameters
        ----------
        teff : float
            effective temperature
        logg : float
            surface gravity
        monh : float
            metallicity
        icor : array_like
            indices of the corner models
        atmo_grid : AtmosphereGrid
            atmosphere grid with input models
        interp : str, optional
            interpolation axis, by default "RHOX"
        itop : int, optional
            index of the topmost point in the RHOX scale, by default 1

        Returns
        -------
        atmo3: Atmosphere
            Interpolated atmosphere
        """

        # We do this for every pair of atmosphere models
        def interpolate(m0, m1, p, param, **kwargs):
            p0 = getattr(m0, param)
            p1 = getattr(m1, param)
            pfrac = (p - p0) / (p1 - p0) if p0 != p1 else 0
            return self.interp_atmo_pair(m0, m1, pfrac, interpvar=interp, **kwargs)

        # Interpolate 8 corner models to create 4 models at the desired [M/H].
        atmo = [[None, None], [None, None]]
        for (i, j) in np.ndindex(2, 2):
            m0 = atmo_grid[icor[0, j, i]]
            m1 = atmo_grid[icor[1, j, i]]
            atmo[i][j] = interpolate(m0, m1, monh, "monh", itop=itop)

        # Interpolate 4 models at the desired [M/H] to create 2 models at desired
        # [M/H] and log(g).
        atmo2 = [None, None]
        for k in range(2):
            atmo2[k] = interpolate(atmo[k][0], atmo[k][1], logg, "logg")

        # Interpolate the 2 models at desired [M/H] and log(g) to create final
        # model at desired [M/H], log(g), and Teff
        atmo3 = interpolate(atmo2[0], atmo2[1], teff, "teff")

        return atmo3

    def spherical_model_correction(self, atmo_grid, icor, logg):
        """
        Correct the radius of the model grids, for the stellar parameters if necessary

        If all interpolated models were spherical, the interpolated model should
        also be reported as spherical. This enables spherical-symmetric radiative
        transfer in the spectral synthesis.

        Formulae for mass and radius at the corners of interpolation cube:
            log(M/Msol) = log g - log g_sol - 2*log(R_sol / R)
            2 * log(R / R_sol) = log g_sol - log g + log(M / M_sol)

        Parameters
        ----------
        atmo_grid : AtmosphereGrid
            complete atmosphere grid
        icor : array_like
            indices for the corner models that were interpolated from the
            atmosphere grid

        Returns
        -------
        geom : {"PP", "SPH"}
            The geometry of the interpolated models
        radius : None, array
            The corrected radius of the models, if geom == "SPH". Otherwise None.
        """

        gtags = atmo_grid.dtype.names
        selection = atmo_grid[icor]
        if "radius" in gtags and "height" in gtags and np.min(selection.radius) > 1:
            mass_cor = (
                selection.logg - logg_sun - 2 * np.log10(R_sun / selection.radius)
            )
            mass = np.mean(mass_cor)
            radius = R_sun * 10 ** ((logg_sun - logg + mass) * 0.5)
            geom = "SPH"
        else:
            geom = "PP"
            radius = None

        return geom, radius

    def interp_atmo_constrained(
        self, x, y, err, par, x2, y2, constraints=None, **kwargs
    ):
        """
        Apply a constraint on each parameter, to have it approach zero

        Parameters
        -------
        x : array[n]
            x data
        y : array[n]
            y data
        err : array[n]
            errors on y data
        par : list[4]
            initial guess for fit parameters - see interp_atmo_func
        x2 : array
            independent variable for tabulated input function
        y2 : array
            dependent variable for tabulated input function
        ** kwargs : dict
            passes keyword arguments to interp_atmo_func

            ndep : int
                number of depth points in supplied quantities
            constraints : array[nconstraint], optional
                error vector for constrained parameters.
                Use errors of 0 for unconstrained parameters.

        Returns
        --------
        ret : list of floats
            best fit parameters
        yfit : array of size (n,)
            best fit to data
        """

        # Evaluate with fixed paramters 3, 4
        # ret = mpfitfun("interp_atmo_func", x, y, err, par, extra=extra, yfit=yfit)
        # TODO: what does the constrain do?
        func = lambda x, *p: self.interp_atmo_func(x, [*p, *par[2:]], x2, y2, **kwargs)
        popt, _ = curve_fit(func, x, y, sigma=err, p0=par[:2])

        ret = [*popt, *par[2:]]
        yfit = func(x, *popt)
        return ret, yfit

    def interp_atmo_func(self, x1, par, x2, y2, ndep=None, y1=None):
        """
        Apply a horizontal shift to x2.
        Apply a vertical shift to y2.
        Interpolate onto x1 the shifted y2 as a function of shifted x2.

        Parameters
        ---------
        x1 : array[ndep1]
            independent variable for output function
        par : array[3]
            shift parameters
            par[0] - horizontal shift for x2
            par[1] - vertical shift for y2
            par[2] - vertical scale factor for y2
        x2 : array[ndep2]
            independent variable for tabulated input function
        y2 : array[ndep2]
            dependent variable for tabulated input function
        ndep : int, optional
            number of depth points in atmospheric structure (default is use all)
        y1 : array[ndep2], optional
            data values being fitted

        Note
        -------
        Only pass y1 if you want to restrict the y-values of extrapolated
        data points. This is useful when solving for the shifts, but should
        not be used when calculating shifted functions for actual use, since
        this restriction can cause discontinuities.
        """
        # Constrained fits may append non-atmospheric quantities to the end of
        # input vector.
        # Extract the output depth scale:
        if ndep is None:
            ndep = len(x1)

        # Shift input x-values.
        # Interpolate input y-values onto output x-values.
        # Shift output y-values.
        x2sh = x2 + par[0]
        y2sh = y2 + par[1]
        y = np.zeros_like(x1)
        y[:ndep] = interp1d(x2sh, y2sh, kind="linear", fill_value="extrapolate")(
            x1[:ndep]
        )  # Note, this implicitly extrapolates

        # Scale output y-values about output y-center.
        ymin = np.min(y[:ndep])
        ymax = np.max(y[:ndep])
        ycen = 0.5 * (ymin + ymax)
        y[:ndep] = ycen + (1.0 + par[2]) * (y[:ndep] - ycen)

        # If extra.y1 was passed, then clip minimum and maximum of output y1.
        if y1 is not None:
            y[:ndep] = np.clip(y[:ndep], np.min(y1), np.max(y1))

        # Set all leftover values to zero
        y[ndep:] = 0
        return y



# -*- coding: utf-8 -*-
from os.path import dirname

import numpy as np
import pytest

from pysme.abund import Abund
from pysme.atmosphere.krzfile import KrzFile
from pysme.linelist.linelist import LineList
from pysme.sme_synth import SME_DLL


# Create Objects to pass to library
# Their functionality is tested in other test files, so we assume it works
@pytest.fixture
def cwd():
    return dirname(__file__)


@pytest.fixture
def libsme():
    return SME_DLL()


@pytest.fixture
def teff():
    return 5000


@pytest.fixture
def grav():
    return 4.2


@pytest.fixture
def vturb():
    return 0.1


@pytest.fixture
def wfirst():
    return 5500


@pytest.fixture
def wlast():
    return 5600


@pytest.fixture
def vw_scale():
    return 2.5


@pytest.fixture
def mu():
    return [1]


@pytest.fixture
def accrt():
    return 0.1


@pytest.fixture
def accwt():
    return 0.1


@pytest.fixture
def linelist():
    linelist = LineList()
    linelist.add("Fe 1", 5502.9931, 0.9582, -3.047, 7.19, -6.22, 239.249)
    linelist.add("Cr 2", 5503.5955, 4.1682, -2.117, 8.37, -6.49, 195.248)
    linelist.add("Mn 1", 5504.0000, 2.0000, -3.000, 8.00, -6.50, 200.247)
    return linelist


@pytest.fixture
def atmo(cwd):
    return KrzFile(cwd + "/testatmo1.krz")


@pytest.fixture
def abund():
    return Abund(monh=0, pattern="Asplund2009")


def test_basic(libsme, wfirst, wlast, vw_scale):
    """Test instantiation of library object and some basic functions"""
    print(libsme.SMELibraryVersion())
    libsme.InputWaveRange(wfirst, wlast)
    libsme.SetVWscale(vw_scale)
    libsme.SetH2broad()

    # assert libsme.file
    # assert libsme.wfirst == wfirst
    # assert libsme.wlast == wlast
    # assert libsme.vw_scale == vw_scale
    # assert libsme.h2broad


def test_linelist(libsme, linelist):
    """Test linelist behaviour"""
    libsme.InputLineList(linelist)
    outlist = libsme.OutputLineList()

    for i in range(len(linelist)):
        outline = [x for x in outlist[i]]
        outline[1] = np.log10(outline[1])

    # TODO
    # libsme.UpdateLineList()

    with pytest.raises(TypeError):
        libsme.InputLineList(None)


def test_atmosphere(libsme, atmo, teff, grav, vturb):
    """Test atmosphere behaviour"""

    libsme.InputModel(teff, grav, vturb, atmo)

    # TODO test different geometries
    with pytest.raises(ValueError):
        libsme.InputModel(-1000, grav, vturb, atmo)

    with pytest.raises(ValueError):
        libsme.InputModel(teff, grav, -10, atmo)

    with pytest.raises(TypeError):
        libsme.InputModel(teff, grav, vturb, None)


def test_abund(libsme, abund):
    """Test abundance behaviour"""
    libsme.InputAbund(abund)

    # TODO: What should be the expected behaviour?
    empty = Abund(monh=0, pattern="empty")
    empty.update_pattern({"H": 12})
    libsme.InputAbund(empty)

    with pytest.raises(TypeError):
        libsme.InputAbund(None)


def test_transf(
    libsme,
    linelist,
    teff,
    grav,
    vturb,
    atmo,
    abund,
    vw_scale,
    wfirst,
    wlast,
    mu,
    accrt,
    accwt,
):
    """Test radiative transfer"""
    libsme.SetLibraryPath()

    libsme.InputLineList(linelist)
    libsme.InputModel(teff, grav, vturb, atmo)
    libsme.InputAbund(abund)
    libsme.Ionization(0)
    libsme.SetVWscale(vw_scale)
    libsme.SetH2broad()

    libsme.InputWaveRange(wfirst, wlast)
    libsme.Opacity()

    nw, wave, synth, cont = libsme.Transf(mu, accrt=accrt, accwi=accwt)
    assert nw == 47

    density = libsme.GetDensity()
    assert np.allclose(density, atmo.rho, rtol=1e-1, equal_nan=True)

    xne = libsme.GetNelec()
    assert np.allclose(xne, atmo.xne, rtol=2e-1)

    xna = libsme.GetNatom()
    assert np.allclose(xna, atmo.xna, rtol=1e-1)

    libsme.GetLineOpacity(linelist.wlcent[0])
    libsme.GetLineRange()
    for switch in [
        "COPSTD",
        "COPRED",
        "COPBLU",
        "AHYD",
        "AH2P",
        "AHMIN",
        "SIGH",
        "AHE1",
        "AHE2",
        "AHEMIN",
        "SIGHE",
        # "ACOOL",
        # "ALUKE",
        "AHOT",
        "SIGEL",
        "SIGH2",
    ]:
        libsme.GetOpacity(switch)



# -*- coding: utf-8 -*-
"""
Utility functions for SME

safe interpolation
"""

import argparse
import builtins
import contextlib
import logging
import os
import subprocess
import sys
from functools import wraps
from platform import python_version

import numpy as np
from numpy import __version__ as npversion
from pandas import __version__ as pdversion
from scipy import __version__ as spversion
from scipy.interpolate import interp1d

from . import __version__ as smeversion
from .sme_synth import SME_DLL

logger = logging.getLogger(__name__)
show_progress_bars = True


def disable_progress_bars():
    global show_progress_bars
    show_progress_bars = False


def enable_progress_bars():
    global show_progress_bars
    show_progress_bars = True


@contextlib.contextmanager
def print_to_log():
    original_print = builtins.print

    def logprint(*args, file=None, **kwargs):
        # The debugger freaks out if we dont give it what it wants
        if file is not None:
            original_print(*args, **kwargs, file=file)
        elif len(args) != 0:
            logger.info(*args, **kwargs)

    builtins.print = logprint
    try:
        yield None
    finally:
        builtins.print = original_print


class getter:
    def __call__(self, func):
        @wraps(func)
        def fget(obj):
            value = func(obj)
            return self.fget(obj, value)

        return fget

    def fget(self, obj, value):
        raise NotImplementedError


class apply(getter):
    def __init__(self, app, allowNone=True):
        self.app = app
        self.allowNone = allowNone

    def fget(self, obj, value):
        if self.allowNone and value is None:
            return value
        if isinstance(self.app, str):
            return getattr(value, self.app)()
        else:
            return self.app(value)


class setter:
    def __call__(self, func):
        @wraps(func)
        def fset(obj, value):
            value = self.fset(obj, value)
            func(obj, value)

        return fset

    def fset(self, obj, value):
        raise NotImplementedError


class oftype(setter):
    def __init__(self, _type, allowNone=True, **kwargs):
        self._type = _type
        self.allowNone = allowNone
        self.kwargs = kwargs

    def fset(self, obj, value):
        if self.allowNone and value is None:
            return value
        elif value is None:
            raise TypeError(
                f"Expected value of type {self._type}, but got None instead"
            )
        return self._type(value, **self.kwargs)


class ofarray(setter):
    def __init__(self, dtype=float, allowNone=True):
        self.dtype = dtype
        self.allowNone = allowNone

    def fset(self, obj, value):
        if self.allowNone and value is None:
            return value
        elif value is None:
            raise TypeError(
                f"Expected value of type {self.dtype}, but got {value} instead"
            )
        arr = np.asarray(value, dtype=self.dtype)
        return np.atleast_1d(arr)


class oneof(setter):
    def __init__(self, allowed_values=()):
        self.allowed_values = allowed_values

    def fset(self, obj, value):
        if value not in self.allowed_values:
            raise ValueError(
                f"Expected one of {self.allowed_values}, but got {value} instead"
            )
        return value


class ofsize(setter):
    def __init__(self, shape, allowNone=True):
        self.shape = shape
        self.allowNone = allowNone
        if hasattr(shape, "__len__"):
            self.ndim = len(shape)
        else:
            self.ndim = 1
            self.shape = (self.shape,)

    def fset(self, obj, value):
        if self.allowNone and value is None:
            return value
        if hasattr(value, "shape"):
            ndim = len(value.shape)
            shape = value.shape
        elif hasattr(value, "__len__"):
            ndim = 1
            shape = (len(value),)
        else:
            ndim = 1
            shape = (1,)

        if ndim != self.ndim:
            raise ValueError(
                f"Expected value with {self.ndim} dimensions, but got {ndim} instead"
            )
        elif not all([i == j for i, j in zip(shape, self.shape)]):
            raise ValueError(
                f"Expected value of shape {self.shape}, but got {shape} instead"
            )
        return value


class absolute(oftype):
    def __init__(self):
        super().__init__(float)

    def fset(self, obj, value):
        value = super().fset(obj, value)
        if value is not None:
            value = abs(value)
        return value


class uppercase(oftype):
    def __init__(self):
        super().__init__(str)

    def fset(self, obj, value):
        value = super().fset(obj, value)
        if value is not None:
            value = value.upper()
        return value


class lowercase(oftype):
    def __init__(self):
        super().__init__(str)

    def fset(self, obj, value):
        value = super().fset(obj, value)
        if value is not None:
            value = value.lower()
        return value


def air2vac(wl_air, copy=True):
    """
    Convert wavelengths in air to vacuum wavelength
    in Angstrom
    Author: Nikolai Piskunov
    """
    if copy:
        wl_vac = np.copy(wl_air)
    else:
        wl_vac = np.asarray(wl_air)
    wl_air = np.asarray(wl_air)

    ii = np.where(wl_air > 1999.352)

    sigma2 = (1e4 / wl_air[ii]) ** 2  # Compute wavenumbers squared
    fact = (
        1e0
        + 8.336624212083e-5
        + 2.408926869968e-2 / (1.301065924522e2 - sigma2)
        + 1.599740894897e-4 / (3.892568793293e1 - sigma2)
    )
    wl_vac[ii] = wl_air[ii] * fact  # Convert to vacuum wavelength

    return wl_vac


def vac2air(wl_vac, copy=True):
    """
    Convert vacuum wavelengths to wavelengths in air
    in Angstrom
    Author: Nikolai Piskunov
    """
    if copy:
        wl_air = np.copy(wl_vac)
    else:
        wl_air = np.asarray(wl_vac)
    wl_vac = np.asarray(wl_vac)

    # Only works for wavelengths above 2000 Angstrom
    ii = np.where(wl_vac > 2e3)

    sigma2 = (1e4 / wl_vac[ii]) ** 2  # Compute wavenumbers squared
    fact = (
        1e0
        + 8.34254e-5
        + 2.406147e-2 / (130e0 - sigma2)
        + 1.5998e-4 / (38.9e0 - sigma2)
    )
    wl_air[ii] = wl_vac[ii] / fact  # Convert to air wavelength
    return wl_air


def safe_interpolation(x_old, y_old, x_new=None, fill_value=0):
    """
    'Safe' interpolation method that should avoid
    the common pitfalls of spline interpolation

    masked arrays are compressed, i.e. only non masked entries are used
    remove NaN input in x_old and y_old
    only unique x values are used, corresponding y values are 'random'
    if all else fails, revert to linear interpolation

    Parameters
    ----------
    x_old : array of size (n,)
        x values of the data
    y_old : array of size (n,)
        y values of the data
    x_new : array of size (m, ) or None, optional
        x values of the interpolated values
        if None will return the interpolator object
        (default: None)

    Returns
    -------
    y_new: array of size (m, ) or interpolator
        if x_new was given, return the interpolated values
        otherwise return the interpolator object
    """

    # Handle masked arrays
    if np.ma.is_masked(x_old):
        x_old = np.ma.compressed(x_old)
        y_old = np.ma.compressed(y_old)

    mask = np.isfinite(x_old) & np.isfinite(y_old)
    x_old = x_old[mask]
    y_old = y_old[mask]

    # avoid duplicate entries in x
    # also sorts data, which allows us to use assume_sorted below
    x_old, index = np.unique(x_old, return_index=True)
    y_old = y_old[index]

    try:
        interpolator = interp1d(
            x_old,
            y_old,
            kind="cubic",
            fill_value=fill_value,
            bounds_error=False,
            assume_sorted=True,
        )
    except ValueError:
        logger.warning(
            "Could not instantiate cubic spline interpolation, using linear instead"
        )
        interpolator = interp1d(
            x_old,
            y_old,
            kind="linear",
            fill_value=fill_value,
            bounds_error=False,
            assume_sorted=True,
        )

    if x_new is not None:
        return interpolator(x_new)
    else:
        return interpolator


def log_version():
    """For Debug purposes"""
    dll = SME_DLL()
    logger.debug("----------------------")
    logger.debug("Python version: %s", python_version())
    try:
        logger.debug("SME CLib version: %s", dll.SMELibraryVersion())
    except OSError:
        logger.debug("SME CLib version: ???")
    logger.debug("PySME version: %s", smeversion)
    logger.debug("Numpy version: %s", npversion)
    logger.debug("Scipy version: %s", spversion)
    logger.debug("Pandas version: %s", pdversion)


def start_logging(
    log_file="log.log",
    level="DEBUG",
    format="%(asctime)-15s - %(levelname)s - %(name)-8s - %(message)s",
):
    """Start logging to log file and command line

    Parameters
    ----------
    log_file : str, optional
        name of the logging file (default: "log.log")
    """

    try:
        level = getattr(logging, str(level).upper())
    except:
        raise ValueError(
            f"Logging level not recognized, try one of ['DEBUG', 'INFO', 'WARNING']"
        )

    name, _ = __name__.split(".", 1)
    logger = logging.getLogger(name)

    logger.setLevel(level)
    filehandler = logging.FileHandler(log_file, mode="w")
    formatter = logging.Formatter(format)
    filehandler.setFormatter(formatter)
    logger.addHandler(filehandler)

    logging.captureWarnings(True)
    log_version()


def redirect_output_to_file(output_file):
    """Redirect ALL output that would go to the commandline, to a file instead

    Parameters
    ----------
    output_file : str
        output filename
    """

    tee = subprocess.Popen(["tee", output_file], stdin=subprocess.PIPE)
    # Cause tee's stdin to get a copy of our stdin/stdout (as well as that
    # of any child processes we spawn)
    os.dup2(tee.stdin.fileno(), sys.stdout.fileno())
    os.dup2(tee.stdin.fileno(), sys.stderr.fileno())

    # The flush flag is needed to guarantee these lines are written before
    # the two spawned /bin/ls processes emit any output
    print("\nHello World", flush=True)
    # print("\nstdout", flush=True)
    # print("stderr", file=sys.stderr, flush=True)

    # These child processes' stdin/stdout are
    # os.spawnve("P_WAIT", "/bin/ls", ["/bin/ls"], {})
    # os.execve("/bin/ls", ["/bin/ls"], os.environ)


def parse_args():
    """Parse command line arguments

    Returns
    -------
    sme : str
        filename to input sme structure
    vald : str
        filename of input linelist or None
    fitparameters : list(str)
        names of the parameters to fit, empty list if none are specified
    """

    parser = argparse.ArgumentParser(description="SME solve")
    parser.add_argument(
        "sme",
        type=str,
        help="an sme input file (either in IDL sav or Numpy npy format)",
    )
    parser.add_argument("--vald", type=str, default=None, help="the vald linelist file")
    parser.add_argument(
        "fitparameters",
        type=str,
        nargs="*",
        help="Parameters to fit, abundances are 'Mg Abund'",
    )
    args = parser.parse_args()
    return args.sme, args.vald, args.fitparameters



# -*- coding: utf-8 -*-
"""
Module for handling linelist data from the VALD3 database (http://vald.astro.uu.se/).


"""
import logging
import re
from io import StringIO
from os.path import dirname, join

import numpy as np
import pandas as pd
import pybtex.database
from astropy import units as u

from ..abund import Abund
from .linelist import LineList, LineListError

logger = logging.getLogger(__name__)


class ValdError(LineListError):
    """Vald Data File Error"""


class ValdFile(LineList):
    """Atomic data for a list of spectral lines."""

    citation_info = r"""
    @ARTICLE{2015PhyS...90e4005R,
        author = {{Ryabchikova}, T. and {Piskunov}, N. and {Kurucz}, R.~L. and
        {Stempels}, H.~C. and {Heiter}, U. and {Pakhomov}, Yu and
        {Barklem}, P.~S.},
        title = "{A major upgrade of the VALD database}",
        journal = {Physica Scripta},
        year = "2015",
        month = "May",
        volume = {90},
        number = {5},
        eid = {054005},
        pages = {054005},
        doi = {10.1088/0031-8949/90/5/054005},
        adsurl = {https://ui.adsabs.harvard.edu/abs/2015PhyS...90e4005R},
        adsnote = {Provided by the SAO/NASA Astrophysics Data System}
    }
    @ARTICLE{2000BaltA...9..590K,
        author = {{Kupka}, F.~G. and {Ryabchikova}, T.~A. and {Piskunov}, N.~E. and
        {Stempels}, H.~C. and {Weiss}, W.~W.},
        title = "{VALD-2 -- The New Vienna Atomic Line Database}",
        journal = {Baltic Astronomy},
        keywords = {ATOMIC DATA, METHODS: SPECTROSCOPIC, STARS: ABUNDANCES, STARS: CHEMICALLY PECULIAR},
        year = "2000",
        month = "Jan",
        volume = {9},
        pages = {590-594},
        doi = {10.1515/astro-2000-0420},
        adsurl = {https://ui.adsabs.harvard.edu/abs/2000BaltA...9..590K},
        adsnote = {Provided by the SAO/NASA Astrophysics Data System}
    }
    @ARTICLE{1999A&AS..138..119K,
        author = {{Kupka}, F. and {Piskunov}, N. and {Ryabchikova}, T.~A. and
        {Stempels}, H.~C. and {Weiss}, W.~W.},
        title = "{VALD-2: Progress of the Vienna Atomic Line Data Base}",
        journal = {\aaps},
        keywords = {ATOMIC DATA, TECHNIQUES: SPECTROSCOPIC, SUN: ABUNDANCES, STARS: ABUNDANCES, STARS: ATMOSPHERES, STARS: CHEMICALLY PECULIAR},
        year = "1999",
        month = "Jul",
        volume = {138},
        pages = {119-133},
        doi = {10.1051/aas:1999267},
        adsurl = {https://ui.adsabs.harvard.edu/abs/1999A&AS..138..119K},
        adsnote = {Provided by the SAO/NASA Astrophysics Data System}
    }
    @ARTICLE{1997BaltA...6..244R,
        author = {{Ryabchikova}, T.~A. and {Piskunov}, N.~E. and {Kupka}, F. and
        {Weiss}, W.~W.},
        title = "{The Vienna Atomic Line Database : Present State and Future Development}",
        journal = {Baltic Astronomy},
        keywords = {DATABASES:  ATOMIC LINE PARAMETERS, STELLAR SPECTROSCOPY},
        year = "1997",
        month = "Mar",
        volume = {6},
        pages = {244-247},
        doi = {10.1515/astro-1997-0216},
        adsurl = {https://ui.adsabs.harvard.edu/abs/1997BaltA...6..244R},
        adsnote = {Provided by the SAO/NASA Astrophysics Data System}
    }
    @ARTICLE{1995A&AS..112..525P,
        author = {{Piskunov}, N.~E. and {Kupka}, F. and {Ryabchikova}, T.~A. and
        {Weiss}, W.~W. and {Jeffery}, C.~S.},
        title = "{VALD: The Vienna Atomic Line Data Base.}",
        journal = {\aaps},
        keywords = {ATOMIC DATA, STARS: ABUNDANCES, ASTRONOMICAL DATA BASES: MISCELLANEOUS},
        year = "1995",
        month = "Sep",
        volume = {112},
        pages = {525},
        adsurl = {https://ui.adsabs.harvard.edu/abs/1995A&AS..112..525P},
        adsnote = {Provided by the SAO/NASA Astrophysics Data System}
    }
    """

    acknowledgement = (
        r"This work has made use of the VALD database, operated at Uppsala University,"
        r"the Institute of Astronomy RAS in Moscow, and the University of Vienna."
    )

    def __init__(self, filename, medium=None):
        self.filename = filename
        self.atmo = None
        self.abund = None
        self.unit = None
        self.energy_unit = None
        linelist = self.loads(filename)

        super().__init__(
            linelist,
            lineformat=self.lineformat,
            medium=self.medium,
            citation_info=self.citation_info,
        )
        # Convert to desired medium
        if medium is not None:
            self.medium = medium

    @staticmethod
    def load(filename):
        """
        Read line data file from the VALD extract stellar service

        Parameters
        ----------
        filename : str
            Name of the VALD linelist file to read

        Returns
        -------
        vald : ValdFile
            Parsed vald file
        """
        return ValdFile(filename)

    def identify_valdtype(self, lines):
        """Determines whether the file was created with extract_all, extract_stellar, or extract_element
        and whether it is in long or short format

        Parameters
        ----------
        lines : list(str)
            file contents
        """
        header = lines[0].strip().split()

        if header[0] == "Damping":
            # short format extract all / extract element
            return "extract_all", "short"
        elif header[0] == "Lande":
            # long format extract all / extract element
            return "extract_all", "long"
        else:
            header = lines[1].strip().split()
            if header[0] == "Damping":
                # short format, extract stellar
                return "extract_stellar", "short"
            elif header[0] == "Lande":
                # long format, extract stellar
                return "extract_stellar", "long"

        raise ValueError("Could not identify ValdFile type")

    def loads(self, filename):
        logger.info("Loading VALD file %s", filename)

        with open(filename, "r") as file:
            lines = file.readlines()

        # Check for Warnings
        while lines[0].lstrip().startswith("WARNING"):
            logger.warning(f"VALD {lines[0].lstrip()}")
            lines = lines[1:]

        # Determine File type and format
        valdtype, fmt = self.identify_valdtype(lines)

        # Determine the number of lines in the file
        if valdtype == "extract_stellar":
            n = self.parse_header(lines[0])
            # Skip the info header if extract stellar
            lines = lines[1:]
        else:
            n = self.parse_nlines(lines, fmt, valdtype)

        # Determine the units and medium in the linelist
        self.parse_columns(lines[1])

        # Split the lines into the different parts
        try:
            if fmt == "long":
                linedata = lines[2 : 2 + n * 4]
                refdata = linedata[3::4]
                if valdtype == "extract_stellar":
                    atmodata = lines[2 + n * 4]
                    abunddata = lines[3 + n * 4 : 21 + n * 4]
            elif fmt == "short":
                linedata = lines[2 : 2 + n]
                refdata = linedata
                if valdtype == "extract_stellar":
                    atmodata = lines[2 + n]
                    abunddata = lines[3 + n : 21 + n]
        except IndexError:
            msg = "Linelist file is shorter than it should be according to the number of lines. Is it incomplete?"
            logger.error(msg)
            raise IOError(msg)

        # Process the individual parts
        linelist = self.parse_linedata(linedata, fmt=fmt, valdtype=valdtype)
        if valdtype == "extract_stellar":
            self.atmo = self.parse_valdatmo(atmodata)
            self.abund = self.parse_abund(abunddata)

        self.citation_info += self.parse_references(refdata, fmt)

        return linelist

    def parse_nlines(self, lines, fmt, valdtype):
        if valdtype == "extract_stellar":
            pattern = r"^ '\S+',$"
        else:
            pattern = r"\* oscillator|References"
        pattern = re.compile(pattern)
        for i in range(len(lines)):
            if re.match(pattern, lines[i]):
                # Offset by the two header lines
                n = i - 2
                break

        if fmt == "long":
            n //= 4
        self.nlines = n
        return self.nlines

    def parse_header(self, line):
        """
        Parse header line from a VALD line data file
        and sets the internal parameters

        Parameters
        ----------
        line : str
            header line of a vald file

        Raises
        ------
        ValdError
            If the header is not understood
        """
        words = [w.strip() for w in line.split(",")]
        # if len(words) < 5 or words[5] != "Wavelength region":
        #     raise ValdError(f"{self.filename} is not a VALD line data file")
        try:
            self.nlines = int(words[2])
            # self._wavelo = float(words[0])
            # self._wavehi = float(words[1])
            # self._nlines_proc = int(words[3])
            # self._vmicro = float(words[4])
            pass
        except:
            raise ValdError(f"{self.filename} is not a VALD line data file")
        return self.nlines

    def parse_columns(self, line):
        match = re.search(r"WL_(air|vac)\((.*?)\)", line)
        medium = match.group(1)
        unit = match.group(2)

        match = re.search(r"E_low\((.*?)\)", line)
        if match is None:
            match = re.search(r"Excit\((.*?)\)", line)
        energy_unit = match.group(1)

        if medium == "air":
            self._medium = "air"
        elif medium == "vac":
            self._medium = "vac"
        else:
            raise ValueError(
                "Could not determine the medium that the wavelength is based on (air or vacuum)"
            )

        if unit == "A":
            self.unit = u.AA
        elif unit == "nm":
            self.unit = u.nm
        elif unit == "cm^-1":
            self.unit = 1 / u.cm
        else:
            raise ValueError("Could not determine the unit of the wavelength")

        if energy_unit == "eV":
            self.energy_unit = u.eV
        elif energy_unit == "cm^-1":
            self.energy_unit = 1 / u.cm
        else:
            raise ValueError("could not determine the unit of the energy levels")

        return self.medium, self.unit, self.energy_unit

    def parse_linedata(self, lines, fmt="short", valdtype="extract_stellar"):
        """Parse line data from a VALD line data file

        Parameters
        ----------
        lines : list of str
            lines of the input data file
        fmt : {"short", "long"}, optional
            linelist format, short format has one
            line of data per spectral line, while the
            long format uses four lines per spectral line.
            The default is "short"

        Returns
        -------
        linelist : LineList
            the parsed linelist
        """

        if fmt == "short":
            names = [
                "species",
                "wlcent",
                "excit",
                "vmic",
                "gflog",
                "gamrad",
                "gamqst",
                "gamvw",
                "lande",
                "reference",
            ]
            if valdtype == "extract_stellar":
                names = names[:-1] + ["depth", names[-1]]

        elif fmt == "long":
            names = [
                "species",
                "wlcent",
                "gflog",
                "excit",
                "j_lo",
                "e_upp",
                "j_up",
                "lande_lower",
                "lande_upper",
                "lande",
                "gamrad",
                "gamqst",
                "gamvw",
            ]
            if valdtype == "extract_stellar":
                names += ["depth"]

            term_lower = lines[1::4]
            term_upper = lines[2::4]
            comment = lines[3::4]
            lines = lines[::4]

        data = StringIO("".join(lines))
        linelist = pd.read_csv(
            data,
            sep=",",
            names=names,
            header=None,
            quotechar="'",
            skipinitialspace=True,
            usecols=range(len(names)),
        )

        # Convert from cm^-1 to eV
        if self.energy_unit == 1 / u.cm:
            conversion_factor = 8065.544
            linelist["excit"] /= conversion_factor
            if fmt == "long":
                linelist["e_upp"] /= conversion_factor

        if fmt == "long":
            comment = [c.replace("'", "").strip() for c in comment]
            linelist["reference"] = comment

            # Parse energy level terms
            # Extract Stellar has quotation marks around the levels
            # extract element does not...
            if valdtype == "extract_stellar":
                term_lower = [t.strip()[8:-1].strip() for t in term_lower]
                term_upper = [t.strip()[8:-1].strip() for t in term_upper]
            else:
                term_lower = [t.strip()[8:].strip() for t in term_lower]
                term_upper = [t.strip()[8:].strip() for t in term_upper]

            term_lower = np.char.partition(term_lower, " ")[:, (0, 2)]
            term_lower = np.char.strip(term_lower)
            idx = term_lower[:, 1] == ""
            term_lower[idx, 1] = term_lower[idx, 0]
            term_lower = np.char.add(
                np.char.add(term_lower[:, 0], " "), term_lower[:, 1]
            )

            term_upper = np.char.partition(term_upper, " ")[:, (0, 2)]
            term_upper = np.char.strip(term_upper)
            idx = term_upper[:, 1] == ""
            term_upper[idx, 1] = term_upper[idx, 0]
            term_upper = np.char.add(
                np.char.add(term_upper[:, 0], " "), term_upper[:, 1]
            )

            linelist["term_lower"] = term_lower
            linelist["term_upper"] = term_upper

            # extract error data
            error = np.array([s[:10].strip() for s in comment])
            error = LineList.parse_line_error(
                error,
                linelist["depth"] if valdtype == "extract_stellar" else None,
            )
            linelist["error"] = error

        # Convert from whatever unit to Angstrom
        factor = self.unit.to(u.AA)
        linelist["wlcent"] *= factor
        self.unit = "Angstrom"
        self.lineformat = fmt

        return linelist

    def parse_valdatmo(self, line):
        """Parse VALD model atmosphere line from a VALD line data file

        Parameters
        ----------
        line : str
            line form the model atmosphere

        Returns
        -------
        atmo : str
            Name of the model atmosphere

        Raises
        ------
        ValdError
            If the line is not from a model atmosphere
        """
        lstr = line.strip()
        if lstr[0] != "'" or lstr[-2:] != "',":
            raise ValdError(f"error parsing model atmosphere: {lstr}")
        return lstr[1:-2]

    def parse_abund(self, lines):
        """Parse VALD abundance lines from a VALD line data file

        Parameters
        ----------
        lines : list of str
            Lines containing the VALD abundance data

        Returns
        -------
        abund : Abund
            Parsed abundance data

        Raises
        ------
        ValdError
            If the data could not be parsed
        """
        abstr = "".join(["".join(line.split()) for line in lines])
        words = [w[1:-1] for w in abstr.split(",")]
        if len(words) != 100 or words[99] != "END":
            raise ValdError(f"Error parsing abundances: {abstr}")
        pattern = [w.split(":") for w in words[:-1]]
        pattern = {el: float(ab) for el, ab in pattern}
        monh = 0
        return Abund(monh, pattern, type="sme")

    def parse_references(self, lines, fmt):
        # Search the linelist data for this pattern, e.g:
        # 1 gf:K14
        # 4 KCN'
        if fmt == "long":
            idiscard = 45
        elif fmt == "short":
            idiscard = 90
        else:
            raise ValueError

        pattern = r"\s\d+ (\w+:)?([\w+]+)[\s']"
        pattern = re.compile(pattern)
        # Discard the initial part of the line
        lines = [l[idiscard:] for l in lines]
        lines = "".join(lines)
        references = [match.group(2) for match in re.finditer(pattern, lines)]
        # We only need each reference ones
        ref = set(references)
        # Multiple references are seperated by '+'
        references = []
        for r in ref:
            references += r.split("+")
        # And make it unique again, if necessary
        references = set(references)
        # some data entries are case sensitive, but bibtex is case insnsitive
        # so remove those and replace them with fixed versions
        if "LWb" in references:
            references.add("LWb2")
            references.remove("LWb")
        if "LGb" in references:
            references.add("LGb2")
            references.remove("LGb")

        # Get references from bibtex file
        # TODO: only load this once? But then again, how often will we do this?
        bibdata = pybtex.database.parse_file(join(dirname(__file__), "VALD3_ref.bib"))
        # DEBUG:
        # pybtex.format_from_string(bibdata.to_string("bibtex"), style="plain", output_backend="plaintext")

        entries = {}
        for r in references:
            try:
                entries[r] = bibdata.entries[r]
            except KeyError as ex:
                logger.warning(f"Could not find citation key: {r}")
                logger.debug(ex)
        bibdata_filtered = pybtex.database.BibliographyData(entries)
        return bibdata_filtered.to_string("bibtex")



# -*- coding: utf-8 -*-
"""
Determine continuum based on continuum mask
and fit best radial velocity to observation
"""

import logging
import warnings

import emcee
import numpy as np
from scipy.constants import speed_of_light
from scipy.interpolate import splev, splrep
from scipy.ndimage import binary_dilation
from scipy.optimize import least_squares
from scipy.signal import correlate
from tqdm import tqdm

from pysme.util import disable_progress_bars

from .iliffe_vector import Iliffe_vector
from .sme import MASK_VALUES
from .sme_synth import SME_DLL
from .util import show_progress_bars

logger = logging.getLogger(__name__)

c_light = speed_of_light * 1e-3  # speed of light in km/s


class ContinuumNormalizationAbstract:
    def __init__(self):
        self.cscale_type = "none"
        pass

    def __call__(self, sme, x_syn, y_syn, segments, rvel=0, least_squares_kwargs=None):
        if least_squares_kwargs is None:
            least_squares_kwargs = {}
        raise NotImplementedError

    def apply(self, wave, smod, cwave, cscale, segments):
        return apply_continuum(wave, smod, cwave, cscale, self.cscale_type, segments)


class ContinuumNormalizationMask(ContinuumNormalizationAbstract):
    def __init__(self):
        super().__init__()
        self.cscale_tyoe = "mask"

    def __call__(
        self,
        sme,
        x_syn,
        y_syn,
        segments,
        rvel=0,
        least_squares_kwargs=None,
    ):
        """
        Fit a polynomial to the spectrum points marked as continuum
        The degree of the polynomial fit is determined by sme.cscale_flag

        Parameters
        ----------
        sme : SME_Struct
            input sme structure with sme.sob, sme.wave, and sme.mask
        segment : int
            index of the wavelength segment to use, or -1 when dealing with the whole spectrum

        Returns
        -------
        cscale : array of size (ndeg + 1,)
            polynomial coefficients of the continuum fit, in numpy order, i.e. largest exponent first
        """

        if segments < 0:
            return sme.cscale

        if least_squares_kwargs is None:
            least_squares_kwargs = {}

        if "spec" not in sme or "wave" not in sme:
            # If there is no observation, we have no continuum scale
            warnings.warn("Missing data for continuum fit")
            cscale = [1]
        elif sme.cscale_flag in ["none", -3]:
            cscale = [1]
        elif sme.cscale_flag in ["fix", -1, -2]:
            # Continuum flag is set to no continuum
            cscale = sme.cscale[segments]
        else:
            # fit a line to the continuum points
            ndeg = sme.cscale_degree

            # Extract points in this segment
            x, y = sme.wave, sme.spec
            if "mask" in sme:
                m = sme.mask
            else:
                m = sme.spec.copy()
                m[:] = sme.mask_value["line"]

            if "uncs" in sme:
                u = sme.uncs
            else:
                u = sme.spec.copy()
                u[:] = 1
            x, y, m, u = x[segments], y[segments], m[segments], u[segments]

            # Set continuum mask
            if np.all((m & MASK_VALUES.CONT) == 0):
                # If no continuum mask has been set
                # Use the effective wavelength ranges of the lines to determine continuum points
                logger.info(
                    "No Continuum mask was set in segment %s, "
                    "Using effective wavelength range of lines to find continuum instead",
                    segments,
                )
                cont = self.get_continuum_mask(x, y, sme.linelist, mask=m)
                # Save mask for next iteration
                m[cont == MASK_VALUES.CONT] = MASK_VALUES.CONT
                logger.debug("Continuum mask points: %i", np.count_nonzero(cont == 2))

            cont = (m & MASK_VALUES.CONT) != 0

            x = x - x[0]
            x, y, u = x[cont], y[cont], u[cont]

            # Fit polynomial
            try:
                func = lambda coef: (np.polyval(coef, x) - y) / u
                c0 = np.polyfit(x, y, deg=ndeg)
                res = least_squares(func, x0=c0, **least_squares_kwargs)
                cscale = res.x
            except TypeError:
                warnings.warn("Could not fit continuum, set continuum mask?")
                cscale = [1]

        return cscale

    def get_continuum_mask(self, wave, synth, linelist, threshold=0.1, mask=None):
        """
        Use the effective wavelength range of the lines,
        to find wavelength points that should be unaffected by lines
        However one usually has to ignore the weak lines, as most points are affected by one line or another
        Therefore keep increasing the threshold until enough lines have been found (>10%)

        Parameters
        ----------
        wave : array of size (n,)
            wavelength points
        linelist : LineList
            LineList object that was input into the Radiative Transfer
        threshold : float, optional
            starting threshold, lines with depth below this value are ignored
            the actual threshold is increased until enough points are found (default: 0.1)

        Returns
        -------
        mask : array(bool) of size (n,)
            True for points between lines and False for points within lines
        """

        if "depth" not in linelist.columns:
            raise ValueError(
                "No depth specified in the linelist, can't auto compute the mask"
            )

        if threshold <= 0:
            threshold = 0.01

        if mask is None:
            mask = np.full(len(wave), MASK_VALUES.LINE)

        # TODO make this better
        dll = SME_DLL()
        dll.linelist = linelist

        width = dll.GetLineRange()
        # TODO: optimize this
        temp = False
        while np.count_nonzero(temp) < len(wave) * 0.1:
            temp = np.full(len(wave), True)
            for i, line in enumerate(width):
                if linelist["depth"][i] > threshold:
                    w = (wave >= line[0]) & (wave <= line[1])
                    temp[w] = False

            # TODO: Good value to increase threshold by?
            temp[mask == MASK_VALUES.BAD] = False
            threshold *= 1.1

        mask[temp] = MASK_VALUES.CONT

        logger.debug("Ignoring lines with depth < %f", threshold)
        return mask


class ContinuumNormalizationMCMC(ContinuumNormalizationAbstract):
    def __init__(self):
        super().__init__()
        self.cscale_tyoe = "mcmc"

    def __call__(
        self,
        sme,
        x_syn,
        y_syn,
        segments,
        rvel=0,
        least_squares_kwargs=None,
    ):
        """
        Fits both radial velocity and continuum level simultaneously
        by comparing the synthetic spectrum to the observation

        The best fit is determined using robust least squares between
        a shifted and scaled synthetic spectrum and the observation

        Parameters
        ----------
        sme : SME_Struct
            contains the observation
        segment : int
            wavelength segment to fit
        x_syn : array of size (ngrid,)
            wavelength of the synthetic spectrum
        y_syn : array of size (ngrid,)
            intensity of the synthetic spectrum

        Returns
        -------
        vrad : float
            radial velocity in km/s
        vrad_unc : float
            radial velocity uncertainty in km/s
        cscale : array of size (ndeg+1,)
            polynomial coefficients of the continuum
        cscale_unc : array if size (ndeg + 1,)
            uncertainties of the continuum coefficients
        """

        if np.isscalar(segments):
            segments = [segments]
        nseg = len(segments)

        if least_squares_kwargs is None:
            least_squares_kwargs = {}

        if sme.cscale_flag in ["none", "fix"] and sme.vrad_flag in [
            "none",
            "fix",
        ]:
            vrad, vunc, cscale, cunc = null_result(nseg, sme.cscale_degree)
            if sme.vrad_flag == "fix":
                vrad = sme.vrad[segments]
            if sme.cscale_flag == "fix":
                cscale = sme.cscale[segments]
            return vrad, vunc, cscale, cunc

        if "spec" not in sme or "wave" not in sme:
            # No observation no radial velocity
            logger.warning("Missing data for radial velocity/continuum determination")
            return null_result(nseg, sme.cscale_degree)

        if "mask" not in sme:
            sme.mask = np.full(sme.spec.size, MASK_VALUES.LINE)
        if "uncs" not in sme:
            sme.uncs = np.full(sme.spec.size, 1.0)

        if np.all(sme.mask_bad[segments].ravel()):
            warnings.warn(
                "Only bad pixels in this segments, can't determine radial velocity/continuum",
                UserWarning,
            )
            return null_result(nseg, sme.cscale_degree)

        if x_syn.ndim == 1:
            x_syn = x_syn[None, :]
        if y_syn.ndim == 1:
            y_syn = y_syn[None, :]

        if x_syn.shape[0] != nseg or y_syn.shape[0] != nseg:
            raise ValueError(
                "Size of synthetic spectrum, does not match the number of requested segments"
            )

        mask = sme.mask_good[segments]
        x_obs = Iliffe_vector(
            [w[m] for w, m in zip(sme.wave[segments].segments, mask.segments)]
        )
        y_obs = sme.spec[segments][mask]
        x_num = x_obs - sme.wave[segments][:, 0]

        if x_obs.size <= sme.cscale_degree:
            warnings.warn(
                "Not enough good pixels to determine radial velocity/continuum"
            )
            return null_result(nseg)

        if sme.cscale_flag in [-3, "none"]:
            cflag = False
            cscale = np.ones((nseg, 1))
            ndeg = 0
        elif sme.cscale_flag in [-1, -2, "fix"]:
            cflag = False
            cscale = sme.cscale[segments]
            ndeg = cscale.shape[1] - 1
        elif sme.cscale_flag in [0, "constant"]:
            ndeg = 0
            cflag = True
        elif sme.cscale_flag in [1, "linear"]:
            ndeg = 1
            cflag = True
        elif sme.cscale_flag in [2, "quadratic"]:
            ndeg = 2
            cflag = True
        else:
            raise ValueError("cscale_flag not recognized")

        if cflag:
            if sme.cscale is not None:
                cscale = sme.cscale[segments]
            else:
                cscale = np.zeros(nseg, ndeg + 1)
                for i, seg in enumerate(segments):
                    cscale[i, -1] = np.nanpercentile(y_obs[seg], 95)

        # Even when the vrad_flag is set to whole
        # you still want to fit the rv of the individual segments
        # just for the continuum fit
        if sme.vrad_flag == "none":
            vrad = np.zeros(len(segments))
            vflag = False
        elif sme.vrad_flag == "whole":
            vrad = sme.vrad[:1]
            vflag = True
        elif sme.vrad_flag == "each":
            vrad = sme.vrad[segments]
            vflag = True
        elif sme.vrad_flag == "fix":
            vrad = sme.vrad[segments]
            vflag = False
        else:
            raise ValueError(f"Radial velocity Flag not understood {sme.vrad_flag}")

        # Limit shift to half an order
        x1, x2 = (
            x_obs[:, 0],
            x_obs[:, [s // 4 for s in x_obs.shape[1]]].ravel(),
        )
        rv_limit = np.abs(c_light * (1 - x2 / x1))
        if sme.vrad_flag == "whole":
            rv_limit = np.min(rv_limit)

        # Use Cross corellatiom as a first guess for the radial velocity
        # This uses the median as the continuum
        if vrad is None:
            y_tmp = np.interp(x_obs, x_syn, y_syn, left=1, right=1)
            corr = correlate(y_obs - np.median(y_obs), y_tmp - 1, mode="same")
            offset = np.argmax(corr)
            x1, x2 = x_obs[offset], x_obs[len(x_obs) // 2]
            vrad = c_light * (1 - x2 / x1)
            if np.abs(vrad) >= rv_limit:
                logger.warning(
                    "Radial Velocity could not be estimated from cross correlation, using initial guess of 0 km/h. Please check results!"
                )
                vrad = 0

        def log_prior(rv, cscale, nwalkers):
            prior = np.zeros(nwalkers)
            # Add the prior here
            # TODO reject too large/small rv values in a prior
            where = np.full(nwalkers, False)
            if vflag:
                where |= np.any(np.abs(rv) > rv_limit, axis=1)
            if cflag:
                where |= np.any(cscale[:, :, -1] < 0, axis=1)
                if ndeg == 1:
                    where |= np.any(
                        (cscale[:, :, -1] + cscale[:, :, -2] * x_num[:, -1]) < 0,
                        axis=1,
                    )
                elif ndeg == 2:
                    for i in range(nseg):
                        where |= np.any(
                            cscale[:, i, None, -1]
                            + cscale[:, i, None, -2] * x_num[i]
                            + cscale[:, i, None, -3] * x_num[i] ** 2
                            < 0,
                            axis=1,
                        )
            prior[where] = -np.inf
            return prior

        def log_prob(par, sep, nseg, ndeg):
            """
            par : array of shape (nwalkers, ndim)
                ndim = 1 for radial velocity + continuum polynomial coeficients
            """
            nwalkers = par.shape[0]
            rv = par[:, :sep] if vflag else vrad[None, :]
            if rv.shape[0] == 1 and nwalkers > 1:
                rv = np.tile(rv, [nwalkers, 1])
            if rv.shape[1] == 1 and nseg > 1:
                rv = np.tile(rv, [1, nseg])
            if cflag:
                cs = par[:, sep:]
                cs.shape = nwalkers, nseg, ndeg + 1
            else:
                cs = cscale[None, ...]

            prior = log_prior(rv, cs, nwalkers)

            # Apply RV shift
            rv_factor = np.sqrt((1 - rv / c_light) / (1 + rv / c_light))
            total = np.zeros(nwalkers)
            for i in range(nseg):
                x = x_obs[i][None, :] * rv_factor[:, i, None]
                model = np.interp(x, x_syn[i], y_syn[i], left=0, right=0)

                # Apply continuum
                y = np.zeros_like(x_num[i])[None, :]
                for j in range(ndeg + 1):
                    y = y * x_num[i] + cs[:, i, j, None]
                model *= y

                # Ignore the non-overlapping parts of the spectrum
                mask = model == 0
                npoints = mask.shape[1] - np.count_nonzero(mask, axis=1)
                resid = (model - y_obs[i]) ** 2
                resid[mask] = 0
                prob = -0.5 * np.sum(resid, axis=-1)
                # Need to rescale here, to account for the ignored points before
                prob *= mask.shape[1] / npoints
                prob[np.isnan(prob)] = -np.inf
                total += prob
            return prior + total

        sep = len(vrad) if vflag else 0
        ndim, p0, scale = 0, [], []
        if vflag:
            ndim += len(vrad)
            p0 += list(vrad)
            scale += [1] * len(vrad)
        if cflag:
            ndim += cscale.size
            p0 += list(cscale.ravel())
            scale += [0.001] * cscale.size
        p0 = np.array(p0)[None, :]
        scale = np.array(scale)[None, :]

        max_n = 10000
        ncheck = 100
        nburn = 300
        nwalkers = max(2 * ndim + 1, 10)
        p0 = p0 + np.random.randn(nwalkers, ndim) * scale
        # If the original guess is good then DEMove is much faster, and sometimes just as good
        # However StretchMove is much more robust to the initial starting value
        moves = [
            (emcee.moves.DEMove(), 0.8),
            (emcee.moves.DESnookerMove(), 0.2),
        ]

        sampler = emcee.EnsembleSampler(
            nwalkers,
            ndim,
            log_prob,
            vectorize=True,
            moves=moves,
            args=(sep, nseg, ndeg),
        )
        # We'll track how the average autocorrelation time estimate changes
        index = 0
        autocorr = np.empty((max_n // ncheck + 1, ndim))
        # This will be useful to testing convergence
        # old_tau = 0

        # Now we'll sample for up to max_n steps
        with tqdm(
            leave=False, desc="RV", total=max_n, disable=~show_progress_bars
        ) as t:
            for _ in sampler.sample(p0, iterations=max_n):
                t.update()
                # Only check convergence every 100 steps
                if sampler.iteration < 2 * nburn or sampler.iteration % ncheck != 0:
                    continue

                # Compute the autocorrelation time so far
                # Using tol=0 means that we'll always get an estimate even
                # if it isn't trustworthy
                tau = sampler.get_autocorr_time(
                    tol=0, discard=sampler.iteration - ncheck
                )
                autocorr[index] = tau
                index += 1

                # Check convergence
                converged = np.all(tau * 100 < sampler.iteration - nburn)
                # converged &= np.all(np.abs(old_tau - tau) < 0.01 * tau)
                # old_tau = tau
                if converged:
                    break

        if sampler.iteration == max_n:
            logger.warning(
                "The radial velocity did not converge within the limit. Check the results!"
            )

        samples = sampler.get_chain(flat=True, discard=nburn)
        _, vrad_unc, _, cscale_unc = null_result(nseg, ndeg)
        if vflag:
            vmin, vrad, vmax = np.percentile(samples[:, :sep], (32, 50, 68), axis=0)
            vrad_unc[:, 0] = vrad - vmin
            vrad_unc[:, 1] = vmax - vrad

        if cflag:
            vmin, cscale, vmax = np.percentile(samples[:, sep:], (32, 50, 68), axis=0)
            vmin.shape = cscale.shape = vmax.shape = nseg, ndeg + 1

            cscale_unc[..., 0] = cscale - vmin
            cscale_unc[..., 1] = vmax - cscale

        if sme.vrad_flag == "whole":
            vrad = np.tile(vrad, [nseg])

        return vrad, vrad_unc, cscale, cscale_unc


class ContinuumNormalizationMatch(ContinuumNormalizationAbstract):
    def __init__(self):
        super().__init__()
        self.cscale_type = "match"
        self.mask = False
        # We over exaggerate the weights on the top of the spectrum
        # this works well to determine the continuum
        # assuming that there is something there
        self.top_factor = 1000
        self.bottom_factor = 1

    def __call__(
        self,
        sme,
        x_syn,
        y_syn,
        segments,
        rvel=0,
        least_squares_kwargs=None,
    ):
        """
        Fit a continuum when no continuum points exist

        Parameters
        ----------
        sme : SME_Struct
            sme structure with observation data
        segment : int
            index of the wavelength segment to fit
        x_syn : array of size (n,)
            wavelengths of the synthetic spectrum
        y_syn : array of size (n,)
            intensity of the synthetic spectrum
        rvel : float, optional
            radial velocity in km/s to apply to the wavelength (default: 0)

        Returns
        -------
        continuum : array of size (ndeg,)
            continuum fit polynomial coefficients
        """

        if sme.cscale_flag == "none":
            return [1]
        elif sme.cscale_flag == "fix":
            return sme.cscale[segments]

        if least_squares_kwargs is None:
            least_squares_kwargs = {}

        # else
        x = sme.wave[segments]
        y = sme.spec[segments]
        u = sme.uncs[segments]

        if self.mask:
            m = sme.mask_cont[segments]
        else:
            m = sme.mask_good[segments]

        # Interpolate the synthetic spectrum to the observed wavelength grid
        yp = apply_radial_velocity([x], [x_syn], [y_syn], [rvel], [0], copy=True)[0]

        # Apply the mask to all vectors
        xs = x - x[0]
        xs, y, u, yp = xs[m], y[m], u[m], yp[m]

        if sme.telluric is not None:
            # by definition sme.telluric is on the same wavelength grid as sme.spec
            # also it should be in the correct restframe
            tell = sme.telluric[segments][m]
            yp *= tell

        # TODO: what should this be?
        if self.top_factor != 0:
            from scipy.ndimage import median_filter

            diff = median_filter(y, 5) - y
            std = 1.5 * np.median(np.abs(diff[diff != 0] - np.median(diff[diff != 0])))
            snr = 1 / std
            factor = 1000 + (snr - 50) * 90
            factor = np.clip(factor, 100, 100_000)
            mod = np.nanmedian(u) / np.nanmedian(y) * (np.nanpercentile(y, 95) - y) ** 2
            u = u + factor * mod

        deg = sme.cscale_degree
        p0 = sme.cscale[segments]

        def func(p):
            model = yp * np.polyval(p, xs)
            resid = (model - y) / u
            # resid[resid > 0] *= self.top_factor
            resid[resid < 0] *= self.bottom_factor
            return resid

        try:
            res = least_squares(
                func,
                x0=p0,
                **least_squares_kwargs,
            )
            popt = res.x
        except (RuntimeError, ValueError) as ex:
            logger.warning("Failed to determine the continuum: " + ex.msg)
            popt = p0

        return popt[None, :]


class ContinuumNormalizationMatchLines(ContinuumNormalizationMatch):
    def __init__(self):
        super().__init__()
        self.cscale_type = "matchlines"
        self.top_factor = 0
        self.bottom_factor = 10


class ContinuumNormalizationMatchMask(ContinuumNormalizationMatch):
    def __init__(self):
        super().__init__()
        self.cscale_type = "match+mask"
        self.mask = True
        self.top_factor = 0


class ContinuumNormalizationMatchLinesMask(ContinuumNormalizationMatchLines):
    def __init__(self):
        super().__init__()
        self.cscale_type = "matchlines+mask"
        self.mask = True


class ContinuumNormalizationSpline(ContinuumNormalizationAbstract):
    def __init__(self):
        super().__init__()
        self.cscale_type = "spline"
        self.mask = False

    def __call__(
        self,
        sme,
        x_syn,
        y_syn,
        segments,
        rvel,
        least_squares_kwargs=None,
    ):
        if sme.cscale_flag in ["none"]:
            return np.ones(len(sme.spec[segments]))
        elif sme.cscale_flag in ["fix"]:
            if sme.cscale is not None:
                return sme.cscale[segments]
            else:
                return np.ones(len(sme.spec[segments]))

        if least_squares_kwargs is None:
            least_squares_kwargs = {}
        least_squares_kwargs.setdefault("f_scale", 0.01)

        w = sme.wave[segments]
        s = sme.spec[segments]
        u = sme.uncs[segments]

        # Apply RV correction to the synthetic spectrum
        y = apply_radial_velocity([w], [x_syn], [y_syn], [rvel], [0], copy=True)[0]
        # and don't forget the telluric spectrum if available
        if sme.telluric is not None:
            tell = sme.telluric[segments]
            y *= tell

        # Apply the bpm to all arrays
        if self.mask:
            m = sme.mask_cont[segments]
        else:
            m = sme.mask_good[segments]

        wm, sm, ym, um = w[m], s[m], y[m], u[m]

        # Fit the spline like in the polynomial
        # so that synth * spline = obs
        func = lambda p: (ym * splev(wm, (t, p, 3)) - sm) / um
        # We use splrep to find the intial guess for the number of knots and their
        # positions
        # TODO: how do we know how many points to use?
        if sme.cscale_flag == "constant":
            tlen = int(np.round(w.max() - w.min()))
        elif sme.cscale_flag == "linear":
            tlen = 3
        elif sme.cscale_flag == "quadratic":
            tlen = 4
        elif sme.cscale_flag == "cubic":
            tlen = 5
        elif sme.cscale_flag == "quintic":
            tlen = 6
        elif sme.cscale_flag == "quantic":
            tlen = 7
        else:
            tlen = int(sme.cscale_flag)
        # We need to avoid the 0 points
        m2 = ym != 0
        # Get a first guess by dividing by ym
        t = np.linspace(wm[m2].min(), wm[m2].max(), tlen)[1:-1]
        t, c, k = splrep(wm[m2], sm[m2] / ym[m2], w=1 / um[m2], k=3, t=t)
        # Then get a real fit using the function
        res = least_squares(
            func,
            x0=c,
            **least_squares_kwargs,
        )
        # And finally evaluate the continuum
        c = res.x
        coef = splev(w, (t, c, k))
        return coef


class ContinuumNormalizationSplineMask(ContinuumNormalizationSpline):
    def __init__(self):
        super().__init__()
        self.cscale_type = "spline+mask"
        self.mask = True


def apply_radial_velocity(wave, wmod, smod, vrad, segments, copy=False):
    if copy:
        wmod = [np.copy(w) for w in wmod]
        smod = [np.copy(s) for s in smod]

    if vrad is None:
        return smod
    for il in segments:
        if vrad[il] is not None:
            rv_factor = np.sqrt((1 + vrad[il] / c_light) / (1 - vrad[il] / c_light))
            wmod[il] *= rv_factor
        smod[il] = np.interp(wave[il], wmod[il], smod[il])
    return smod


def apply_continuum(wave, smod, cwave, cscale, cscale_type, segments, copy=False):
    if copy:
        smod = np.copy(smod)

    if cscale is None:
        return smod
    for il in segments:
        if cscale[il] is not None and not np.all(cscale[il] == 0):
            if cscale_type in ["spline", "spline+mask"]:
                if len(cscale[il]) != len(smod[il]):
                    cs = np.interp(wave[il], cwave[il], cscale[il])
                else:
                    cs = cscale[il]
                smod[il] *= cs
            else:
                x = wave[il] - wave[il][0]
                smod[il] *= np.polyval(cscale[il], x)
    return smod


def apply_radial_velocity_and_continuum(
    wave, wmod, smod, vrad, cscale, cscale_type, segments, copy=False
):
    """
    Apply the radial velocity and continuum corrections
    to a syntheic spectrum to match the observation

    Parameters
    ----------
    wave : array
        final wavelength array of the observation
    wmod : array
        wavelength array of the synthethic spectrum
    smod : array
        flux array of the synthetic spectrum
    vrad : array, None
        radial velocities in km/s for each segment
    cscale : array, None
        continnum scales for each segment, exact meaning depends on cscale_type
    cscale_type : str
        defines the continuum correction behaviour
    segments : array
        the segments to apply the correction to

    Returns
    -------
    smod : array
        the corrected synthetic spectrum
    """
    smod = apply_radial_velocity(wave, wmod, smod, vrad, segments, copy=copy)
    # The radial velocity shift also interpolates onto the wavelength grid
    smod = apply_continuum(wave, smod, wave, cscale, cscale_type, segments, copy=copy)
    return smod


def null_result(nseg, ndeg=0, ctype=None):
    vrad, vrad_unc = np.zeros(nseg), np.zeros((nseg, 2))
    if ctype in ["spline", "spline+mask"]:
        cscale = [np.ones(ndeg[i]) for i in range(nseg)]
        cscale = Iliffe_vector(values=cscale)
        cscale_unc = [np.zeros(ndeg[i]) for i in range(nseg)]
        cscale_unc = Iliffe_vector(values=cscale_unc)
    else:
        cscale, cscale_unc = np.zeros((nseg, ndeg + 1)), np.zeros((nseg, ndeg + 1, 2))
        cscale[:, -1] = 1
    return vrad, vrad_unc, cscale, cscale_unc


def cross_correlate_segment(x_obs, y_obs, x_syn, y_syn, mask, mask_wider, rv_bounds):
    # Interpolate synthetic observation onto the observation wavelength grid
    y_tmp = np.interp(x_obs, x_syn, y_syn)

    # Normalize both spectra
    y_obs_tmp = np.copy(y_obs)
    y_obs_tmp /= np.nanpercentile(y_obs_tmp, 95)
    y_obs_tmp -= 1
    y_obs_tmp[~mask] = 0

    y_tmp_tmp = np.copy(y_tmp)
    y_tmp_tmp /= np.nanpercentile(y_tmp_tmp, 95)
    y_tmp_tmp -= 1
    mask_wider = np.interp(x_obs, x_syn, mask_wider) > 0.5
    y_tmp_tmp[~mask_wider] = 0

    # Perform cross correaltion between normalized spectra
    corr = correlate(y_obs_tmp, y_tmp_tmp, mode="same")

    # Determine the radial velocity offset
    # and only retain the area within the bounds
    x_mid = x_obs[len(x_obs) // 2]
    x_shift = c_light * (1 - x_mid / x_obs)

    idx = (x_shift >= rv_bounds[0]) & (x_shift <= rv_bounds[1])
    x_shift = x_shift[idx]
    corr = corr[idx]

    return x_shift, corr


def determine_radial_velocity(
    sme,
    x_syn,
    y_syn,
    segment,
    cscale=None,
    whole=False,
    least_squares_kwargs=None,
):
    """
    Calculate radial velocity by using cross correlation and
    least-squares between observation and synthetic spectrum

    Parameters
    ----------
    sme : SME_Struct
        sme structure with observed spectrum and flags
    segment : int
        which wavelength segment to handle, -1 if its using the whole spectrum
    cscale : array of size (ndeg,)
        continuum coefficients, as determined by e.g. determine_continuum
    x_syn : array of size (n,)
        wavelength of the synthetic spectrum
    y_syn : array of size (n,)
        intensity of the synthetic spectrum

    Raises
    ------
    ValueError
        if sme.vrad_flag is not recognized

    Returns
    -------
    rvel : float
        best fit radial velocity for this segment/whole spectrum
        or None if no observation is present
    """

    if least_squares_kwargs is None:
        least_squares_kwargs = {}
    least_squares_kwargs.setdefault("bounds", (-100, 100))
    least_squares_kwargs.setdefault("jac", "3-point")
    least_squares_kwargs.setdefault("loss", "soft_l1")
    least_squares_kwargs.setdefault("method", "dogbox")
    least_squares_kwargs.setdefault("x_scale", "jac")
    least_squares_kwargs.setdefault("ftol", 1e-8)
    least_squares_kwargs.setdefault("xtol", 1e-8)
    least_squares_kwargs.setdefault("gtol", 1e-8)

    if "spec" not in sme or "wave" not in sme:
        # No observation no radial velocity
        warnings.warn("Missing data for radial velocity determination")
        rvel = 0
    elif sme.vrad_flag == "none":
        # vrad_flag says don't determine radial velocity
        rvel = sme.vrad[segment]
    elif sme.vrad_flag == "whole" and not whole:
        # We are inside a segment, but only want to determine rv at the end
        rvel = 0
    elif sme.vrad_flag == "fix":
        rvel = sme.vrad[segment]
    else:
        # Fit radial velocity
        # Extract data
        x, y = sme.wave, sme.spec
        if "mask" in sme:
            m = sme.mask
        else:
            m = Iliffe_vector(
                [np.full(s, MASK_VALUES.LINE, dtype="i4") for s in sme.spec.shape[1]]
            )

        if "uncs" in sme:
            u = sme.uncs
        else:
            u = sme.spec.copy()
            u[:] = 1

        # Only this one segment
        x_obs = x[segment]
        y_obs = y[segment].copy()
        u_obs = u[segment]
        mask = m[segment]
        if sme.telluric is not None:
            tell = sme.telluric[segment]
        else:
            tell = 1

        if sme.vrad_flag == "each":
            # apply continuum
            y_syn = apply_continuum(
                {segment: x_syn},
                {segment: y_syn},
                sme.wave,
                {segment: cscale},
                sme.cscale_type,
                [segment],
            )[segment]
        elif sme.vrad_flag == "whole":
            # All segments
            y_syn = apply_continuum(
                x_syn,
                y_syn,
                sme.wave,
                cscale,
                sme.cscale_type,
                range(len(y_obs)),
            )
        else:
            raise ValueError(
                f"Radial velocity flag {sme.vrad_flag} not recognised, expected one of 'each', 'whole', 'none'"
            )

        # Filter the mask for the correct sections
        # Need to use temporary m instead of mask, so that we can check
        m = (mask & MASK_VALUES.VRAD) != 0
        if not np.any(m):
            # No VRAD specified, use LINE instead
            m = (mask & MASK_VALUES.LINE) != 0
        mask = m
        mask &= u_obs != 0

        # Widen the mask by roughly the amount expected from the rv_bounds
        bounds = least_squares_kwargs["bounds"]
        rv = max(bounds)
        rv_factor = np.sqrt((1 + rv / c_light) / (1 - rv / c_light))
        # mask_wider = mask.copy()
        if sme.vrad_flag == "each":
            iterations = int(
                np.ceil(np.nanmedian((x_obs * rv_factor - x_obs) / np.gradient(x_obs)))
            )
            iterations = max(1, iterations)
            mask_wider = binary_dilation(mask, iterations=iterations)
            mask_wider = np.interp(x_syn, x_obs, mask_wider) > 0.5
        elif sme.vrad_flag == "whole":
            mask_wider = [None] * len(x_obs)
            for i in range(len(x_obs)):
                iterations = int(
                    np.ceil(
                        np.nanmedian(
                            (x_obs[i] * rv_factor - x_obs[i]) / np.gradient(x_obs[i])
                        )
                    )
                )
                iterations = max(1, iterations)
                mask_wider[i] = binary_dilation(mask[i], iterations=iterations)
                mask_wider[i] = np.interp(x_syn[i], x_obs[i], mask_wider[i]) > 0.5
            mask_wider = Iliffe_vector(mask_wider)
        else:
            raise ValueError

        # Get a first rough estimate from cross correlation
        if sme.vrad_flag == "each":
            x_shift, corr = cross_correlate_segment(
                x_obs, y_obs, x_syn, y_syn, mask, mask_wider, bounds
            )
        else:
            # If using several segments we run the cross correlation for each
            # segment seperately and then sum the results
            nseg = len(segment)
            shift = [None for _ in range(nseg)]
            corr = [None for _ in range(nseg)]
            for i in range(len(x_obs)):
                shift[i], corr[i] = cross_correlate_segment(
                    x_obs[i],
                    y_obs[i],
                    x_syn[i],
                    y_syn[i],
                    mask[i],
                    mask_wider[i],
                    bounds,
                )

            n_min = min(len(s) for s in shift)
            x_shift = np.linspace(bounds[0], bounds[1], n_min)
            corrs_interp = [np.interp(x_shift, s, c) for s, c in zip(shift, corr)]
            corrs_interp = np.array(corrs_interp)
            corr = np.sum(corrs_interp, axis=0)

            # Concatenate the segments for the last step
            mask = np.concatenate(mask)
            mask_wider = np.concatenate(mask_wider)
            x_obs = np.concatenate(x_obs)
            y_obs = np.concatenate(y_obs)
            u_obs = np.concatenate(u_obs)
            x_syn = np.concatenate(x_syn)
            y_syn = np.concatenate(y_syn)
            if sme.telluric is not None:
                tell = np.concatenate(tell)

        # Retrieve the initial cross correlation guess
        offset = np.argmax(corr)
        rvel = x_shift[offset]

        # Normalize the spectra to 1
        y_obs = y_obs / np.nanpercentile(y_obs, 95)
        y_obs -= 1
        y_syn = y_syn / np.nanpercentile(y_syn, 95)
        y_syn -= 1

        # Apply mask
        y_obs[~mask] = 0
        y_syn[~mask_wider] = 0
        u_obs = u_obs.copy()
        u_obs[~mask] = 1

        # Then minimize the least squares for a better fit
        # as cross correlation can only find
        def func(rv):
            rv_factor = np.sqrt((1 - rv / c_light) / (1 + rv / c_light))
            shifted = interpolator(x_obs * rv_factor)
            resid = (y_obs - shifted * tell) / u_obs
            resid[~mask] = 0
            resid = np.nan_to_num(resid, copy=False, nan=0, posinf=0, neginf=0)
            return resid

        interpolator = lambda x: np.interp(x, x_syn, y_syn)
        try:
            res = least_squares(func, x0=rvel, **least_squares_kwargs)
            rvel = res.x[0]
        except ValueError:
            logger.warning(f"Could not determine radial velocity for segment {segment}")
    return rvel


def match_rv_continuum(sme, segments, x_syn, y_syn):
    """
    Match both the continuum and the radial velocity of observed/synthetic spectrum

    Note that the parameterization of the continuum is different to old SME !!!

    Parameters
    ----------
    sme : SME_Struct
        input sme structure with all the parameters
    segment : int
        index of the wavelength segment to match, or -1 when dealing with the whole spectrum
    x_syn : array of size (n,)
        wavelength of the synthetic spectrum
    y_syn : array of size (n,)
        intensitz of the synthetic spectrum

    Returns
    -------
    rvel : float
        new radial velocity
    cscale : array of size (ndeg + 1,)
        new continuum coefficients
    """

    cont_func = {
        "mask": ContinuumNormalizationMask,
        "match": ContinuumNormalizationMatch,
        "match+mask": ContinuumNormalizationMatchMask,
        "matchlines": ContinuumNormalizationMatchLines,
        "matchlines+mask": ContinuumNormalizationMatchLinesMask,
        "spline": ContinuumNormalizationSpline,
        "spline+mask": ContinuumNormalizationSplineMask,
        "mcmc": ContinuumNormalizationMCMC,
    }

    vrad, vrad_unc, cscale, cscale_unc = null_result(
        sme.nseg, sme.cscale_degree, sme.cscale_type
    )
    if sme.cscale_flag == "none" and sme.vrad_flag == "none":
        return cscale, cscale_unc, vrad, vrad_unc

    if np.isscalar(segments):
        segments = [segments]

    if not callable(sme.vrad_flag):
        radial_velocity = determine_radial_velocity
    else:
        radial_velocity = sme.vrad_flag

    if not callable(sme.cscale_type):
        continuum_normalization = cont_func[sme.cscale_type]()
    else:
        continuum_normalization = sme.cscale_type

    least_squares_kwargs = dict(
        bounds=sme.vrad_bounds,
        loss=sme.vrad_loss,
        method=sme.vrad_method,
        jac=sme.vrad_jac,
        x_scale=sme.vrad_xscale,
        ftol=sme.vrad_ftol,
        xtol=sme.vrad_xtol,
        gtol=sme.vrad_gtol,
    )

    if sme.vrad_flag == "none":
        pass
    elif sme.vrad_flag == "fix":
        vrad[segments] = sme.vrad[segments]
    elif sme.vrad_flag == "each":
        for s in segments:
            # We only use the continuum mask for the continuum fit,
            # we need the lines for the radial velocity
            vrad[s] = radial_velocity(
                sme,
                x_syn[s],
                y_syn[s],
                s,
                cscale[s],
                whole=False,
                least_squares_kwargs=least_squares_kwargs,
            )
    elif sme.vrad_flag == "whole":
        s = segments
        vrad[s] = radial_velocity(
            sme,
            [x_syn[s] for s in s],
            [y_syn[s] for s in s],
            s,
            cscale[s],
            whole=True,
            least_squares_kwargs=least_squares_kwargs,
        )
    else:
        raise ValueError

    if sme.cscale_flag == "none":
        pass
    elif sme.cscale_flag == "fix":
        if sme.cscale is not None:
            cscale[segments] = sme.cscale[segments]
        else:
            pass
    else:
        if sme.cscale_type == "mcmc":
            for s in segments:
                (
                    vrad[s],
                    vrad_unc[s],
                    cscale[s],
                    cscale_unc[s],
                ) = continuum_normalization(
                    sme,
                    x_syn[s],
                    y_syn[s],
                    s,
                    rvel=vrad[s],
                )
        else:
            for s in segments:
                cscale[s] = continuum_normalization(
                    sme,
                    x_syn[s],
                    y_syn[s],
                    s,
                    rvel=vrad[s],
                    least_squares_kwargs=dict(
                        bounds=sme.cscale_bounds,
                        loss=sme.cscale_loss,
                        method=sme.cscale_method,
                        jac=sme.cscale_jac,
                        x_scale=sme.cscale_xscale,
                        ftol=sme.cscale_ftol,
                        xtol=sme.cscale_xtol,
                        gtol=sme.cscale_gtol,
                    ),
                )

    # Keep values from unused segments
    select = np.arange(sme.nseg)
    mask = np.full(select.shape, True)
    for seg in segments:
        mask &= select != seg
    vrad[mask] = sme.vrad[mask]
    if sme.cscale_type in ["spline", "spline+mask"]:
        for i in range(len(mask)):
            if (
                mask[i]
                and sme.cscale is not None
                and len(cscale[i]) == len(sme.cscale[i])
            ):
                cscale[i] = sme.cscale[i]
    else:
        cscale[mask] = sme.cscale[mask]

    return cscale, cscale_unc, vrad, vrad_unc



# -*- coding: utf-8 -*-
"""
Defines the colors to use in plotting
"""


class PlotColors:
    """Define the colors to use in plotting"""

    def __init__(self):
        self.fmt = {
            "Obs": {"color": "#1f77b4", "linestyle": "solid"},
            "Syn": {"color": "#ff7f0e", "linestyle": "solid", "marker": ""},
            "Orig": {"color": "#2ca02c", "linestyle": "dashed"},
            "LineMask": {"facecolor": "#bcbd22", "alpha": 1},
            "ContMask": {"facecolor": "#d62728", "alpha": 1},
        }
        self.plotly_styles = {
            "solid": None,
            "dashed": "dash",
            "dashdot": "dashdot",
            "dotted": "dot",
        }

    def __getitem__(self, key):
        return self.fmt[key]



# -*- coding: utf-8 -*-
r"""
Wrapper for IDL style C libary code
``{return_type} {func_name}(int argv, void *argc[]);``

with argv - number of parameters
and argc - list of pointers to those parameters
"""
import ctypes as ct
import logging
import platform
import warnings

import numpy as np

from .smelib.libtools import load_library

logger = logging.getLogger(__name__)


class IDL_String(ct.Structure):
    """
    IDL strings are actually structures like this one,
    for correct passing of values we need to define this structure
    NOTE: the definition might have changed with IDL version,
    so make sure to use the same one as in the C code
    """

    _fields_ = [("slen", ct.c_int), ("stype", ct.c_ushort), ("s", ct.c_char_p)]


MAX_ELEM = 100
MOSIZE = 288
MUSIZE = 77
MAX_PATHLEN = 512
MAX_OUT_LEN = 511
SP_LEN = 8


class GlobalState(ct.Structure):
    _fields_ = [
        ("NRHOX", ct.c_short),
        ("NRHOX_allocated", ct.c_short),
        ("MOTYPE", ct.c_short),
        ("TEFF", ct.c_double),
        ("GRAV", ct.c_double),
        ("WLSTD", ct.c_double),
        ("RADIUS", ct.c_double),
        ("NumberSpectralSegments", ct.c_int),
        ("NLINES", ct.c_int),
        ("NWAVE_C", ct.c_int),
        ("WFIRST", ct.c_double),
        ("WLAST", ct.c_double),
        ("VW_SCALE", ct.c_double),
        ("N_SPLIST", ct.c_int),
        ("IXH1", ct.c_int),
        ("IXH2", ct.c_int),
        ("IXH2mol", ct.c_int),
        ("IXH2pl", ct.c_int),
        ("IXHMIN", ct.c_int),
        ("IXHE1", ct.c_int),
        ("IXHE2", ct.c_int),
        ("IXHE3", ct.c_int),
        ("IXC1", ct.c_int),
        ("IXAL1", ct.c_int),
        ("IXSI1", ct.c_int),
        ("IXSI2", ct.c_int),
        ("IXCA1", ct.c_int),
        ("IXMG1", ct.c_int),
        ("IXMG2", ct.c_int),
        ("IXCA2", ct.c_int),
        ("IXN1", ct.c_int),
        ("IXFE1", ct.c_int),
        ("IXO1", ct.c_int),
        ("IXCH", ct.c_int),
        ("IXNH", ct.c_int),
        ("IXOH", ct.c_int),
        ("PATHLEN", ct.c_int),
        ("change_byte_order", ct.c_int),
        ("allocated_NLTE_lines", ct.c_int),
        ("flagMODEL", ct.c_short),
        ("flagWLRANGE", ct.c_short),
        ("flagABUND", ct.c_short),
        ("flagLINELIST", ct.c_short),
        ("flagIONIZ", ct.c_short),
        ("flagCONTIN", ct.c_short),
        ("lineOPACITIES", ct.c_short),
        ("flagH2broad", ct.c_short),
        ("initNLTE", ct.c_short),
        ("IFOP", ct.c_short * 20),
        ("ABUND", ct.c_float * MAX_ELEM),
        ("RHOX", ct.c_double * MOSIZE),
        ("T", ct.c_double * MOSIZE),
        ("XNE", ct.c_double * MOSIZE),
        ("XNA", ct.c_double * MOSIZE),
        ("RHO", ct.c_double * MOSIZE),
        ("VTURB", ct.c_double * MOSIZE),
        ("RAD_ATMO", ct.c_double * MOSIZE),
        ("XNA_eos", ct.c_double * MOSIZE),
        ("XNE_eos", ct.c_double * MOSIZE),
        ("RHO_eos", ct.c_double * MOSIZE),
        ("AHYD", ct.c_double * MOSIZE),
        ("AH2P", ct.c_double * MOSIZE),
        ("AHMIN", ct.c_double * MOSIZE),
        ("SIGH", ct.c_double * MOSIZE),
        ("AHE1", ct.c_double * MOSIZE),
        ("AHE2", ct.c_double * MOSIZE),
        ("AHEMIN", ct.c_double * MOSIZE),
        ("SIGHE", ct.c_double * MOSIZE),
        ("ACOOL", ct.c_double * MOSIZE),
        ("ALUKE", ct.c_double * MOSIZE),
        ("AHOT", ct.c_double * MOSIZE),
        ("SIGEL", ct.c_double * MOSIZE),
        ("SIGH2", ct.c_double * MOSIZE),
        ("TKEV", ct.c_double * MOSIZE),
        ("TK", ct.c_double * MOSIZE),
        ("HKT", ct.c_double * MOSIZE),
        ("TLOG", ct.c_double * MOSIZE),
        ("FREQ", ct.c_double),
        ("FREQLG", ct.c_double),
        ("EHVKT", ct.c_double * MOSIZE),
        ("STIM", ct.c_double * MOSIZE),
        ("BNU", ct.c_double * MOSIZE),
        ("H1FRACT", ct.c_float * MOSIZE),
        ("HE1FRACT", ct.c_float * MOSIZE),
        ("H2molFRACT", ct.c_float * MOSIZE),
        ("COPBLU", ct.c_double * MOSIZE),
        ("COPRED", ct.c_double * MOSIZE),
        ("COPSTD", ct.c_double * MOSIZE),
        ("LINEOP", ct.POINTER(ct.c_double) * MOSIZE),
        ("AVOIGT", ct.POINTER(ct.c_double) * MOSIZE),
        ("VVOIGT", ct.POINTER(ct.c_double) * MOSIZE),
        ("LTE_b", ct.c_double * MOSIZE),
        ("PATH", ct.c_char * MAX_PATHLEN),
        ("debug_print", ct.c_int),
        ("ATOTAL", ct.POINTER(ct.POINTER(ct.c_double))),
        ("INDX_C", ct.POINTER(ct.c_int)),
        ("YABUND", ct.POINTER(ct.c_double)),
        ("XMASS", ct.POINTER(ct.c_double)),
        ("EXCUP", ct.POINTER(ct.c_double)),
        ("ENU4", ct.POINTER(ct.c_double)),
        ("ENL4", ct.POINTER(ct.c_double)),
        ("BNLTE_low", ct.POINTER(ct.POINTER(ct.c_double))),
        ("BNLTE_upp", ct.POINTER(ct.POINTER(ct.c_double))),
        ("FRACT", ct.POINTER(ct.POINTER(ct.c_float))),
        ("PARTITION_FUNCTIONS", ct.POINTER(ct.POINTER(ct.c_float))),
        ("POTION", ct.POINTER(ct.c_float)),
        ("MOLWEIGHT", ct.POINTER(ct.c_float)),
        ("MARK", ct.POINTER(ct.c_short)),
        ("AUTOION", ct.POINTER(ct.c_short)),
        ("IDLHEL", ct.POINTER(ct.c_short)),
        ("ION", ct.POINTER(ct.c_int)),
        ("ANSTEE", ct.POINTER(ct.c_int)),
        ("WLCENT", ct.POINTER(ct.c_double)),
        ("EXCIT", ct.POINTER(ct.c_double)),
        ("GF", ct.POINTER(ct.c_double)),
        ("GAMRAD", ct.POINTER(ct.c_double)),
        ("GAMQST", ct.POINTER(ct.c_double)),
        ("GAMVW", ct.POINTER(ct.c_double)),
        ("ALMAX", ct.POINTER(ct.c_double)),
        ("Wlim_left", ct.POINTER(ct.c_double)),
        ("Wlim_right", ct.POINTER(ct.c_double)),
        ("SPLIST", ct.c_char_p),
        ("spname", ct.c_char_p),
        ("SPINDEX", ct.POINTER(ct.c_int)),
        ("flagNLTE", ct.POINTER(ct.c_short)),
        ("result", ct.c_char * (MAX_OUT_LEN + 1)),
    ]

    def free_linelist(self):
        if self.flagLINELIST:
            self.spname = None
            self.SPINDEX = None
            self.ION = None
            self.MARK = None
            self.AUTOION = None
            self.WLCENT = None
            self.EXCIT = None
            self.GF = None
            self.GAMRAD = None
            self.GAMQST = None
            self.GAMVW = None
            self.ANSTEE = None
            self.IDLHEL = None
            self.ALMAX = None
            self.Wlim_left = None
            self.Wlim_right = None
            self.flagLINELIST = 0

    def free_ionization(self):
        if self.flagIONIZ:
            self.SPLIST = None
            for i in range(self.NRHOX_allocated):
                self.FRACT[i] = None
                self.PARTITION_FUNCTIONS[i] = None
            self.FRACT = None
            self.PARTITION_FUNCTIONS = None
            self.POTION = None
            self.MOLWEIGHT = None
            self.flagIONIZ = 0

    def free_opacities(self):
        if self.lineOPACITIES:
            for i in range(self.NRHOX):
                self.LINEOP[i] = None
                self.AVOIGT[i] = None
                self.VVOIGT[i] = None
            self.lineOPACITIES = 0


def get_lib_name():
    """Get the name of the sme C library"""
    system = platform.system().lower()
    arch = platform.machine()
    bits = 64  # platform.architecture()[0][:-3]

    return "sme_synth.so.{system}.{arch}.{bits}".format(
        system=system, arch=arch, bits=bits
    )


def get_typenames(arg):
    """
    Return internal typename based on the type of the argument
    strings -> "unicode"
    floating points -> "double"
    integers -> "int"
    """
    if isinstance(arg, (str, np.str)) or (
        isinstance(arg, np.ndarray) and np.issubdtype(arg.dtype, np.str)
    ):
        return "unicode"
    if isinstance(arg, (float, np.floating)) or (
        isinstance(arg, np.ndarray) and np.issubdtype(arg.dtype, np.floating)
    ):
        return "double"
    if isinstance(arg, (int, np.integer)) or (
        isinstance(arg, np.ndarray) and np.issubdtype(arg.dtype, np.integer)
    ):
        return "int"

    raise ValueError("argument datatype not understood")


def get_dtype(type):
    """
    Get the ctypes dtype appropiate for the passed type string

    Parameters
    -------
    type : str
        One of 'int', 'short', 'long', 'float', 'double',
        'unicode' or one of the first letters 'islfdu'
    Returns
    ------
    type: class
        corresponding ctypes type
    """
    if type in ["i", "int", int]:
        return ct.c_int
    elif type in ["s", "short"]:
        return ct.c_short
    elif type in ["l", "long"]:
        return ct.c_long
    elif type in ["f", "float"]:
        return ct.c_float
    elif type in ["d", "double", float]:
        return ct.c_double
    elif type in ["u", "unicode", "str", str]:
        return IDL_String
    elif type in ["state", GlobalState]:
        return ct.POINTER(GlobalState)
    else:
        raise ValueError("Data type {type} not understood".format(type=type))


def is_nullptr(ptr):
    try:
        # If we can access it without an exception its not a null ptr
        tmp = ptr[0]
        return False
    except ValueError:
        return True


def get_c_dtype(ptr):
    if isinstance(ptr, bytes):
        return ct.c_char
    while hasattr(ptr, "contents"):
        ptr = ptr._type_
    return ptr


def get_c_shape(field, ptr, state):
    if isinstance(ptr, bytes):
        return (len(ptr),)
    if isinstance(ptr, ct._Pointer):
        if isinstance(ptr[0], ct._Pointer):
            size1 = state.contents.NRHOX
            size2 = state.contents.N_SPLIST
            return (size1, size2)
        else:
            size = state.contents.NLINES
            return (size,)
    return None


def idl_call_external(funcname, *args, restype="str", type=None, lib=None, state=None):
    r"""
    Call an external C library (here the SME lib) function that uses the IDL type interface
    i.e. restype func(int n, void \*args[]), where n is the number of arguments, and args is a list of pointers to the arguments

    Input arrays will be converted to the required datatype for the C function (see type keyword),
    and any changes to input arrays will be written back if possible.
    Input arrays that are already in the correct datatype will not be copied (and the values can therefore change in the C call)

    Note that all strings are converted into IDL_String objects, even those that are in arrays

    Parameters
    ----------
    funcname : str
        Name of the function to call in the library
    restype : str, optional
        expected type of the return value (default: "str")
    type : str, list(str), optional
        type of the input parameters, will default to int/double for all integer/floating point values.
        Accepted values are ('short', 'int', 'long', 'float', 'double', 'unicode') or their respective first letters.
        This means one can use a string as shorthand, e.g. "iidds"

    Returns
    -------
    value : restype
        return value of the function call
    """

    # Load library if that wasn't done yet
    if lib is None:
        lib = load_library()

    # prepare input arguments
    args = list(args)
    staying_alive = [a for a in args]
    original = [a for a in args]

    # datatype is determined by passed type keyword
    # defaults to 'int' for all integer type values and 'double' for all floating point values
    if type is None:
        type = [get_typenames(a) for a in args]
    elif type in ["short", "int", "long", "float", "double"]:
        type = [type for i in range(len(args))]

    # Parse arguments into c values
    # keep Python elements in staying alive, so they are not discarded by the garbage collection
    for i in range(len(args)):
        # Single values
        if isinstance(args[i], (int, float, np.number)):
            dtype = get_dtype(type[i])
            staying_alive[i] = np.array(args[i]).astype(dtype, copy=False).ctypes
            args[i] = staying_alive[i].data
        elif isinstance(args[i], str):
            staying_alive[i] = IDL_String(
                slen=len(args[i]), stype=1, s=args[i].encode()
            )
            args[i] = ct.addressof(staying_alive[i])
        # Arrays
        elif isinstance(args[i], (list, np.ndarray)):
            if isinstance(args[i], list):
                # enforce numpy arrays
                args[i] = np.array(args[i])
            if np.issubdtype(args[i].dtype, np.number) or np.issubdtype(
                args[i].dtype, np.bool_
            ):
                dtype = get_dtype(type[i])
                args[i] = np.require(
                    args[i], dtype=dtype, requirements=["C", "A", "W", "O"]
                )
                staying_alive[i] = args[i].ctypes
                args[i] = staying_alive[i].data
            elif np.issubdtype(args[i].dtype, np.str_) or np.issubdtype(
                args[i].dtype, np.bytes_
            ):
                args[i] = args[i].astype("S")
                staying_alive.append(args[i])
                length = [len(a) for a in args[i]]
                args[i] = [
                    IDL_String(slen=l, stype=1, s=s) for s, l in zip(args[i], length)
                ]
                staying_alive.append(args[i])

                strarr = (IDL_String * len(args[i]))()
                for j in range(len(args[i])):
                    strarr[j] = args[i][j]

                staying_alive[i] = strarr
                args[i] = ct.addressof(strarr)
            else:
                raise TypeError("Array datatype not understood")
        else:
            raise TypeError("Argument type not understood")

    # Load function and define parameters
    func = getattr(lib, funcname)
    if state is not None:
        func.argtypes = (
            ct.c_int,
            ct.POINTER(ct.c_void_p),
            ct.POINTER(GlobalState),
        )
    else:
        func.argtypes = (ct.c_int, ct.POINTER(ct.c_void_p))

    if restype in ["str", str]:
        func.restype = ct.c_char_p
    else:
        func.restype = get_dtype(restype)

    # Convert input parameters to list of void pointers
    a = np.array(args, dtype=ct.c_void_p)
    a = np.ascontiguousarray(a)

    argc = len(args)
    argv = a.ctypes.data_as(ct.POINTER(ct.c_void_p))

    # C function call
    if state is not None:
        res = func(argc, argv, state)
    else:
        res = func(argc, argv)

    # Try to copy back data to the original array memory (if necessary)
    for i in range(len(original)):
        if isinstance(original[i], np.ndarray):
            if np.issubdtype(original[i].dtype, np.number) or np.issubdtype(
                original[i].dtype, np.bool_
            ):
                # do nothing if its the same array
                if original[i] is staying_alive[i]._arr:
                    continue
                arr = staying_alive[i]._arr
            elif np.issubdtype(original[i].dtype, np.str_):
                # For string arrays recover the strings from the IDL_String structure
                arr = [s.s.decode() for s in staying_alive[i]]
            else:
                # Shouldn't happen
                continue

            # If nothing was changed then all is good
            if not np.all(original[i] == arr):
                try:
                    original[i][:] = arr
                except ValueError as ve:
                    print(
                        "WARNING: Array values changed, but could not be written back to the original array\n{ve}".format(
                            ve=str(ve)
                        )
                    )

    return res


class IDL_DLL:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
            IDL_DLL.init(cls._instance, *args, **kwargs)
        return cls._instance

    @staticmethod
    def init(self, libfile=None):
        self.libfile = libfile
        self.lib = load_library(libfile)

        # Pick best interface
        self.interfaces = self.get_interfaces()
        if "Parallel" in self.interfaces:
            self.interface = "Parallel"
        elif "IDL" in self.interfaces:
            self.interface = "IDL"
        else:
            self.interface = self.interfaces[0]

    def __getattr__(self, name):
        return lambda *args, **kwargs: self.call(name, *args, **kwargs)

    def call(self, name, *args, raise_error=True, raise_warning=False, **kwargs):
        """
        run idl_call_external and check for errors in the output

        Parameters
        ----------
        name : str
            name of the external C function to call
        args
            parameters for the function
        kwargs
            keywords for the function

        Raises
        --------
        ValueError
            If the returned string is not empty, it means an error occured in the C library
        """
        error = ""
        try:
            error = idl_call_external(
                self.get_name(name), *args, lib=self.lib, **kwargs
            )
        except AttributeError as ex:
            error = "Using obsolete SME Library; {ex}".format(ex=ex)
            raise_error = False
            raise_warning = True

        if error != b"":
            if hasattr(error, "decode"):
                error = error.decode()

            if raise_error:
                raise ValueError(
                    "{name} (call external): {error}".format(name=name, error=error)
                )
            if raise_warning:
                warnings.warn(
                    "{name} (call external): {error}".format(name=name, error=error)
                )
        return error

    def get_name(self, funname):
        if self.interface == "Parallel":
            return f"Parallel_{funname}"
        elif self.interface == "IDL":
            return funname
        elif self.interface == "Cython":
            return f"Cython_{funname}"
        else:
            raise ValueError

    def new_state(self):
        # ctypes initializes the fields to 0/NULL by default
        state = GlobalState()
        return ct.pointer(state)

    def free_state(self, state):
        del state

    def copy_state(self, state):
        new = GlobalState()
        new.NRHOX = state.contents.NRHOX
        new.NRHOX_allocated = state.contents.NRHOX_allocated
        new.MOTYPE = state.contents.MOTYPE
        new.TEFF = state.contents.TEFF
        new.GRAV = state.contents.GRAV
        new.WLSTD = state.contents.WLSTD
        new.RADIUS = state.contents.RADIUS
        new.NumberSpectralSegments = state.contents.NumberSpectralSegments
        new.NLINES = state.contents.NLINES
        new.NWAVE_C = state.contents.NWAVE_C
        new.WFIRST = state.contents.WFIRST
        new.WLAST = state.contents.WLAST
        new.VW_SCALE = state.contents.VW_SCALE
        new.N_SPLIST = state.contents.N_SPLIST
        new.IXH1 = state.contents.IXH1
        new.IXH2 = state.contents.IXH2
        new.IXH2mol = state.contents.IXH2mol
        new.IXH2pl = state.contents.IXH2pl
        new.IXHMIN = state.contents.IXHMIN
        new.IXHE1 = state.contents.IXHE1
        new.IXHE2 = state.contents.IXHE2
        new.IXHE3 = state.contents.IXHE3
        new.IXC1 = state.contents.IXC1
        new.IXAL1 = state.contents.IXAL1
        new.IXSI1 = state.contents.IXSI1
        new.IXSI2 = state.contents.IXSI2
        new.IXCA1 = state.contents.IXCA1
        new.IXMG1 = state.contents.IXMG1
        new.IXMG2 = state.contents.IXMG2
        new.IXCA2 = state.contents.IXCA2
        new.IXN1 = state.contents.IXN1
        new.IXFE1 = state.contents.IXFE1
        new.IXO1 = state.contents.IXO1
        new.IXCH = state.contents.IXCH
        new.IXNH = state.contents.IXNH
        new.IXOH = state.contents.IXOH
        new.PATHLEN = state.contents.PATHLEN
        new.change_byte_order = state.contents.change_byte_order
        new.allocated_NLTE_lines = state.contents.allocated_NLTE_lines
        new.flagMODEL = state.contents.flagMODEL
        new.flagWLRANGE = state.contents.flagWLRANGE
        new.flagABUND = state.contents.flagABUND
        new.flagLINELIST = state.contents.flagLINELIST
        new.flagIONIZ = state.contents.flagIONIZ
        new.flagCONTIN = state.contents.flagCONTIN
        new.lineOPACITIES = state.contents.lineOPACITIES
        new.flagH2broad = state.contents.flagH2broad
        new.initNLTE = state.contents.initNLTE
        new.FREQ = state.contents.FREQ
        new.FREQLG = state.contents.FREQLG
        new.debug_print = state.contents.debug_print

        ct.memmove(new.IFOP, state.contents.IFOP, ct.sizeof(new.IFOP))
        ct.memmove(new.ABUND, state.contents.ABUND, ct.sizeof(new.ABUND))
        ct.memmove(new.RHOX, state.contents.RHOX, ct.sizeof(new.RHOX))
        ct.memmove(new.T, state.contents.T, ct.sizeof(new.T))
        ct.memmove(new.XNE, state.contents.XNE, ct.sizeof(new.XNE))
        ct.memmove(new.XNA, state.contents.XNA, ct.sizeof(new.XNA))
        ct.memmove(new.RHO, state.contents.RHO, ct.sizeof(new.RHO))
        ct.memmove(new.VTURB, state.contents.VTURB, ct.sizeof(new.VTURB))
        ct.memmove(new.RAD_ATMO, state.contents.RAD_ATMO, ct.sizeof(new.RAD_ATMO))
        ct.memmove(new.XNA_eos, state.contents.XNA_eos, ct.sizeof(new.XNA_eos))
        ct.memmove(new.XNE_eos, state.contents.XNE_eos, ct.sizeof(new.XNE_eos))
        ct.memmove(new.RHO_eos, state.contents.RHO_eos, ct.sizeof(new.RHO_eos))
        ct.memmove(new.AHYD, state.contents.AHYD, ct.sizeof(new.AHYD))
        ct.memmove(new.AH2P, state.contents.AH2P, ct.sizeof(new.AH2P))
        ct.memmove(new.AHMIN, state.contents.AHMIN, ct.sizeof(new.AHMIN))
        ct.memmove(new.SIGH, state.contents.SIGH, ct.sizeof(new.SIGH))
        ct.memmove(new.AHE1, state.contents.AHE1, ct.sizeof(new.AHE1))
        ct.memmove(new.AHE2, state.contents.AHE2, ct.sizeof(new.AHE2))
        ct.memmove(new.AHEMIN, state.contents.AHEMIN, ct.sizeof(new.AHEMIN))
        ct.memmove(new.SIGHE, state.contents.SIGHE, ct.sizeof(new.SIGHE))
        ct.memmove(new.ACOOL, state.contents.ACOOL, ct.sizeof(new.ACOOL))
        ct.memmove(new.ALUKE, state.contents.ALUKE, ct.sizeof(new.ALUKE))
        ct.memmove(new.AHOT, state.contents.AHOT, ct.sizeof(new.AHOT))
        ct.memmove(new.SIGEL, state.contents.SIGEL, ct.sizeof(new.SIGEL))
        ct.memmove(new.SIGH2, state.contents.SIGH2, ct.sizeof(new.SIGH2))
        ct.memmove(new.TKEV, state.contents.TKEV, ct.sizeof(new.TKEV))
        ct.memmove(new.TK, state.contents.TK, ct.sizeof(new.TK))
        ct.memmove(new.HKT, state.contents.HKT, ct.sizeof(new.HKT))
        ct.memmove(new.TLOG, state.contents.TLOG, ct.sizeof(new.TLOG))
        ct.memmove(new.EHVKT, state.contents.EHVKT, ct.sizeof(new.EHVKT))
        ct.memmove(new.STIM, state.contents.STIM, ct.sizeof(new.STIM))
        ct.memmove(new.BNU, state.contents.BNU, ct.sizeof(new.BNU))
        ct.memmove(new.H1FRACT, state.contents.H1FRACT, ct.sizeof(new.H1FRACT))
        ct.memmove(new.HE1FRACT, state.contents.HE1FRACT, ct.sizeof(new.HE1FRACT))
        ct.memmove(
            new.H2molFRACT,
            state.contents.H2molFRACT,
            ct.sizeof(new.H2molFRACT),
        )
        ct.memmove(new.COPBLU, state.contents.COPBLU, ct.sizeof(new.COPBLU))
        ct.memmove(new.COPRED, state.contents.COPRED, ct.sizeof(new.COPRED))
        ct.memmove(new.COPSTD, state.contents.COPSTD, ct.sizeof(new.COPSTD))
        ct.memmove(new.LTE_b, state.contents.LTE_b, ct.sizeof(new.LTE_b))

        nrhox = state.contents.NRHOX
        nsplist = state.contents.N_SPLIST
        nlines = state.contents.NLINES

        def ct_copy(new, state, field, shape, dtype=None):
            value = getattr(state.contents, field)
            dtype = get_c_dtype(value) if dtype is None else dtype
            if is_nullptr(value):
                return None
            if len(shape) == 1:
                copy = (dtype * shape[0])()
                ct.memmove(copy, value, shape[0] * ct.sizeof(dtype))
            if len(shape) == 2:
                copy = (ct.POINTER(dtype) * shape[0])()
                for i in range(shape[0]):
                    copy[i] = (dtype * shape[1])()
                    ct.memmove(copy[i], value[i], shape[1] * ct.sizeof(dtype))
            return copy

        for i in range(nrhox):
            new.LINEOP[i] = (ct.c_double * nlines)()
            ct.memmove(
                new.LINEOP[i],
                state.contents.LINEOP[i],
                nlines * ct.sizeof(ct.c_double),
            )
            new.AVOIGT[i] = (ct.c_double * nlines)()
            ct.memmove(
                new.AVOIGT[i],
                state.contents.AVOIGT[i],
                nlines * ct.sizeof(ct.c_double),
            )
            new.VVOIGT[i] = (ct.c_double * nlines)()
            ct.memmove(
                new.VVOIGT[i],
                state.contents.VVOIGT[i],
                nlines * ct.sizeof(ct.c_double),
            )

        new.ATOTAL = ct_copy(new, state, "ATOTAL", (nrhox, nsplist), ct.c_double)
        new.INDX_C = ct_copy(new, state, "INDX_C", (nlines,), ct.c_int)
        new.YABUND = ct_copy(new, state, "YABUND", (nlines,), ct.c_double)
        new.XMASS = ct_copy(new, state, "XMASS", (nlines,), ct.c_double)
        new.EXCUP = ct_copy(new, state, "EXCUP", (nlines,), ct.c_double)
        new.ENU4 = ct_copy(new, state, "ENU4", (nlines,), ct.c_double)
        new.ENL4 = ct_copy(new, state, "ENL4", (nlines,), ct.c_double)
        new.BNLTE_low = ct_copy(new, state, "BNLTE_low", (nlines, nrhox), ct.c_double)
        new.BNLTE_upp = ct_copy(new, state, "BNLTE_upp", (nlines, nrhox), ct.c_double)
        new.FRACT = ct_copy(new, state, "FRACT", (nrhox, nsplist), ct.c_float)
        new.PARTITION_FUNCTIONS = ct_copy(
            new, state, "PARTITION_FUNCTIONS", (nrhox, nsplist), ct.c_float
        )
        new.POTION = ct_copy(new, state, "POTION", (nsplist,), ct.c_float)
        new.MOLWEIGHT = ct_copy(new, state, "MOLWEIGHT", (nsplist,), ct.c_float)
        new.MARK = ct_copy(new, state, "MARK", (nlines,), ct.c_short)
        new.AUTOION = ct_copy(new, state, "AUTOION", (nlines,), ct.c_short)
        new.IDLHEL = ct_copy(new, state, "IDLHEL", (nlines,), ct.c_short)
        new.ION = ct_copy(new, state, "ION", (nlines,), ct.c_int)
        new.ANSTEE = ct_copy(new, state, "ANSTEE", (nlines,), ct.c_int)
        new.WLCENT = ct_copy(new, state, "WLCENT", (nlines,), ct.c_double)
        new.EXCIT = ct_copy(new, state, "EXCIT", (nlines,), ct.c_double)
        new.GF = ct_copy(new, state, "GF", (nlines,), ct.c_double)
        new.GAMRAD = ct_copy(new, state, "GAMRAD", (nlines,), ct.c_double)
        new.GAMQST = ct_copy(new, state, "GAMQST", (nlines,), ct.c_double)
        new.GAMVW = ct_copy(new, state, "GAMVW", (nlines,), ct.c_double)
        new.ALMAX = ct_copy(new, state, "ALMAX", (nlines,), ct.c_double)
        new.Wlim_left = ct_copy(new, state, "Wlim_left", (nlines,), ct.c_double)
        new.Wlim_right = ct_copy(new, state, "Wlim_right", (nlines,), ct.c_double)
        new.SPINDEX = ct_copy(new, state, "SPINDEX", (nlines,), ct.c_int)
        new.flagNLTE = ct_copy(new, state, "flagNLTE", (nlines,), ct.c_short)
        new.SPLIST = state.contents.SPLIST
        new.spname = state.contents.spname
        new.PATH = state.contents.PATH
        new.result = state.contents.result

        return ct.pointer(new)

    def get_interfaces(self):
        try:
            interfaces = idl_call_external("GetInterfaces")
            interfaces = interfaces.decode()
            interfaces = interfaces.split(";")
        except AttributeError:
            # Old Library without that function
            interfaces = ["IDL"]
        return interfaces



# -*- coding: utf-8 -*-
from collections import OrderedDict

import numpy as np
import pytest

from pysme.abund import Abund

pattern_names = ["Asplund2009", "Grevesse2007", "Empty"]
types = ["H=12", "n/nH", "n/nTot", "SME"]


def test_init_with_too_few_args():
    """Test that __init__ raise an error if too few arguments are passed."""
    Abund()
    Abund(monh=0)
    Abund(pattern="asplund2009")
    Abund(type="H=12")


def test_init_using_pattern_names():
    """Test handling of abundance pattern name passed to __init__()."""
    # Each abundance pattern name yields an Abund object.
    for pattern_name in pattern_names:
        assert isinstance(Abund(pattern=pattern_name, monh=0), Abund)

    # The 'Empty' abundance pattern has a value of None for all elements.
    abund = Abund(monh=0, pattern="Empty")
    pattern = abund.get_pattern(raw=True)
    assert pattern[0] == 0
    assert np.all(np.isnan(pattern[1:]))

    # An invalid abundance pattern name raises an error.
    with pytest.raises(ValueError):
        Abund(monh=0, pattern="INVALID")


def test_call_returns_abund_in_odict():
    """Test return value, which is an ordered dictionary with element
    abbreviations as the keys and abundances as the values.
    """
    abund = Abund(pattern=pattern_names[0], monh=0)
    assert isinstance(abund(), dict)
    assert tuple(abund().keys()) == abund.elem


def test_getitem_returns_abund_values():
    """Test getitem method, which return computed abundance values for
    the specified element or list of elements.
    """
    abund = Abund(pattern=pattern_names[0], monh=0)
    assert abund["H"] == 12


def test_monh_property_set_and_get():
    """Test setting and getting monh property. Set converts input to float."""
    # Input str convertable to float yields a float with the specified value.
    abund = Abund(pattern=pattern_names[0], monh="-6e-1")
    assert isinstance(abund.monh, float)
    assert abund.monh == -0.6

    # Input int yields a float with the specified value.
    abund.monh = -2
    assert isinstance(abund.monh, float)
    assert abund.monh == -2.0

    # Input float yields a float with the specified value.
    abund.monh = 0.3
    assert isinstance(abund.monh, float)
    assert abund.monh == 0.3

    # Input str that cannot be converted to float raises an error.
    with pytest.raises(ValueError):
        abund = Abund(pattern=pattern_names[0], monh="ABC")

    # Input that is not a string or a number raises an error.
    with pytest.raises(TypeError):
        abund.monh = []


def test_pattern_property_set_and_get():
    """Test setting and getting pattern property. Set is not allowed."""
    # Raise error is user tries to set pattern
    abund = Abund(pattern="Empty", monh=0)
    with pytest.raises(AttributeError):
        abund.pattern = 0.0


def test_update_pattern():
    """Test behavior of update_pattern(), which modifies values in _pattern
    for the specified element(s).
    """
    # Update for one element yields float with the specified value.
    abund = Abund(pattern="Empty", monh=0)
    assert np.isnan(abund["Fe"])
    abund.update_pattern({"Fe": "3.14"})
    assert isinstance(abund["Fe"], float)
    assert abund["Fe"] == 3.14

    # Update for two elements yields floats with the specified values.
    abund.update_pattern({"C": 8.4, "F": 5})
    assert isinstance(abund["C"], float)
    assert isinstance(abund["F"], float)
    assert abund["C"] == 8.4
    assert abund["F"] == 5.0


def test_totype_fromtype():
    """Test behavior of totype() and fromtype(), which are static methods
    that return a copy of the input abundance pattern transformed to or
    from the specified abudnance pattern type.
    """
    # Round trip tests that compare copy=fromtype(totype()) with original.
    orig = Abund(pattern=pattern_names[0], monh=0)()
    for type in types:
        pattern = Abund.totype(orig, type)
        copy = Abund.fromtype(pattern, type)
        # Same elements in the same order for full dictionary.
        assert copy.keys() == orig.keys()
        # Same elements have abundance defined (!= None).
        o = OrderedDict((k, v) for k, v in orig.items() if not np.isnan(v))
        c = OrderedDict((k, v) for k, v in copy.items() if not np.isnan(v))
        assert c.keys() == o.keys()
        # Logarithmic abundances differ by less than 1e-10.
        assert all([abs(o[k] - c[k]) < 1e-10 for k in o.keys()])
        # Lowercase type yields same result as mixed case type.
        # type_lc = type.lower()
        # assert copy == Abund.fromtype(Abund.totype(orig, type_lc), type_lc)

    # Invalid abundance pattern type raises error.
    with pytest.raises(ValueError):
        copy = Abund.totype(orig, "INVALID")
    with pytest.raises(ValueError):
        copy = Abund.fromtype(orig, "INVALID")
